package NTLSystem;

import java.awt.Component;
import java.util.Vector;
import NTLSystem.features.*;

/**
 * An interface representing an editable data element to be stored in
 * an EditableDataDictionary.
 * <P>
 * EditableDataElements have the ability to:
 * <UL>
 * <LI>Get the key of the data element.  This is usually it's name,
 * which is unique in the system or collection of elements of the same
 * type.
 * <LI>View the element, returning a Component which contains the data in
 * the object in a read-only manner.
 * <LI>Edit the element, returning a Component which contains the data in
 * the object in an editable manner.
 * </UL>
 * <P>
 * Objects which implement EditableDataElement are <EM>strongly</EM>
 * encouraged to implement Cloneable as well.
 * <P>
 * $Id: EditableDataElement.java,v 1.6 1998/10/16 01:56:01 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       EditableDataDictionary
 * @version   $Revision: 1.6 $
 **/
public interface EditableDataElement
extends Cloneable
{

  /**
   * EditableDataElements must implement Object.Clone()
   **/
  public Object clone();

  /**
   * Returns the key for this element.  This is usually the unique
   * name of the data element.
   *
   * @return     A String key for this element.
   **/
  public String getKey();

  /**
   * Returns a Panel which contains the GUI functionality necessary
   * for viewing the object.  Viewing can be viewed as a special case of
   * read-only editing.  However, the panel should not appear as an
   * editing panel wherever possible [text strings should not be able
   * to be changed, etc].
   *
   * @return    A Panel which contains the functionality to view this
   *            object.
   **/
  public Component getViewPanel();

  /**
   * Returns a Panel which contains the GUI functionality necessary
   * for editing the object.  This panel should make copies of the
   * data in the variables, and only save the values back into the
   * object when <A HREF="#saveValues()">saveValues()</A> is called.
   *
   * @return    A Panel which contains the functionality for editing
   *            this object.
   **/
  public Component getEditPanel();

  /**
   * Saves the values in the currently-being-edited panel back into
   * the object. 
   *
   * @exception NotBeingEditedException If this method is called while
   *            no editing is going on.
   * @exception CannotSaveValuesException If the values cannot be
   *            saved [the reason is stated in the exception
   *            message]. 
   **/
  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException;

  /**
   * Method to test if the values currently in the panel returned by
   * <A HREF="#Edit()">Edit()</A> can be saved back to the object.
   * This method should perform any consistency checking necessary,
   * based on the data in the Panel.  It should return a StringBuffer
   * stating why the fields cannot be saved, or null indicating that
   * they can. 
   *
   * @param     reasons A Vector to fill with reasons [as Strings] why
   *            the values cannot be saved.
   * @exception NotBeingEditedException If this EditableDataElement is
   *            not currently being edited.
   **/
  public void canSaveValues(Vector reasons)
       throws NotBeingEditedException;

  /**
   * Provides a way for EditableDataElements to release editing
   * resources.
   **/
  public void doneEditing();

}
