package NTLSystem;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import NTLSystem.support.WindowCloser;

/**
 * An object which allows the addition of elements from an
 * EditableDataDictionary to an EditableDataContainer.  The object is
 * responsible for taking care of the addition and removal of data
 * elements, and for any type and consistency checking.
 * <P>
 * $Id: EditableDataManipulator.java,v 1.4 1998/07/31 21:05:29 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       EditableDataDictionary
 * @see       EditableDataContainer
 * @version   $Revision: 1.4 $
 **/
public class EditableDataManipulator
extends JFrame
implements ActionListener
{

  /**
   * The EditableDataContainer to callback to when changes take
   * place.
   **/
  private EditableDataContainer edc;
  /**
   * The EditableDataDictionary to get possible values from.
   **/
  private EditableDataDictionary edd;
  /**
   * The list of existing elements [from the EditableDataContainer].
   **/
  private JList existingList;
  /**
   * The list of possible elements [from the EditableDataDictionary].
   **/
  private JList possibleList;
  /**
   * A ListModel for EditableDataContainers.
   **/
  private EDCListModel edclm;


  /**
   * The constructor.
   *
   * @param     frameName The title of the frame.
   * @param     edc The EditableDataContainer from which values are
   *            being added and removed.
   * @param     edd The EditableDataDictionary from which possible
   *            value names are being determined.
   **/
  public EditableDataManipulator(String frameName,
				 EditableDataContainer edc,
				 EditableDataDictionary edd)
  {
    super(frameName);
    this.edc = edc;
    this.edd = edd;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    JButton b;

    getContentPane().setLayout(gbl);
    setSize(500, 400);

    // Add the existing list, the add button and the possibilities
    // list.
    gbc.fill = GridBagConstraints.BOTH;
    gbc.gridwidth = 1;
    edclm = new EDCListModel(edc);
    existingList = new JList(edclm);
    existingList.setVisibleRowCount(10);
    existingList.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
    existingList.setFixedCellWidth(150);

    JScrollPane jsp = new JScrollPane(existingList);
    gbl.setConstraints(jsp, gbc);
    getContentPane().add(jsp);

    gbc.fill = GridBagConstraints.NONE;
    gbc.gridwidth = GridBagConstraints.RELATIVE;
    gbc.weightx = gbc.weighty = 0.0;
    b = new JButton("< Add <");
    b.setActionCommand("ADD");
    b.addActionListener(this);
    b.setBorder(BorderFactory.
		createCompoundBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2),
				     b.getBorder()));
    gbl.setConstraints(b, gbc);
    getContentPane().add(b);

    gbc.fill = GridBagConstraints.BOTH;
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    possibleList = new JList( edd.dict );
    possibleList.setVisibleRowCount(10);
    possibleList.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
    possibleList.setFixedCellWidth(150);
    jsp = new JScrollPane(possibleList);
    gbl.setConstraints(jsp, gbc);
    getContentPane().add(jsp);

    // Add the buttons at the bottom

    gbc.fill = GridBagConstraints.HORIZONTAL;
    gbc.gridwidth = 1;
    b = new JButton("Remove Selected");
    b.setActionCommand("REMOVE");
    b.addActionListener(this);
    b.setBorder(BorderFactory.
		createCompoundBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2),
				     b.getBorder()));
    gbl.setConstraints(b, gbc);
    getContentPane().add(b);

    gbc.anchor = GridBagConstraints.CENTER;
    gbc.gridwidth = GridBagConstraints.RELATIVE;
    b = new JButton("Dismiss");
    b.setActionCommand("DISMISS");
    b.addActionListener(this);
    b.setBorder(BorderFactory.
		createCompoundBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2),
				     b.getBorder()));
    gbl.setConstraints(b, gbc);
    getContentPane().add(b);

    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbc.fill = GridBagConstraints.HORIZONTAL;
    b = new JButton("Edit Dictionary");
    b.setActionCommand("EDIT_DICT");
    b.addActionListener(this);
    b.setBorder(BorderFactory.
		createCompoundBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2),
				     b.getBorder()));
    gbl.setConstraints(b, gbc);
    getContentPane().add(b);

    // Add a window closer.
    // This will automagically attach itself as a WindowListener...
    new WindowCloser(this);

    pack();
    setVisible(true);
  }

  /**
   * Handles ActionEvents for the frame.
   *
   * @see java.awt.event.ActionListener
   **/
  public void actionPerformed(ActionEvent ae)
  {
    String acmd = ae.getActionCommand();
    if (acmd.equals("ADD")) {

      // Add selected elements from the possible list to the
      // container.
      Object toAdd[] = possibleList.getSelectedValues();
      for (int i=0; i<toAdd.length; i++) {
	edc.addElement( (String)toAdd[i] );
      } // end for over i
      // Update the existing list.
      edclm.update();
    } else if (acmd.equals("REMOVE")) {

      // Remove the selected elements from the container.
      Object toRemove[] = existingList.getSelectedValues();
      for (int i=0; i<toRemove.length; i++) {
	edc.removeElement( (String)toRemove[i] );
      } // end for over i
      // Update the existing list.
      edclm.update();
    } else if (acmd.equals("EDIT_DICT")) {

      // Edit the dictionary
      Thread t = new Thread() {
	public void run() {
	  // Wait until done
	  edd.Edit(true);
	  // The possible list will automagically be updated because
	  // the JList uses the EditableDataDictionary's ListModel
	}
      };
      t.start();

    } else if (acmd.equals("DISMISS")) {

      // Close the window.
      dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));

    }
  }

  /* INNER CLASS for creating a JList from an EditableDataContainer. */
  class EDCListModel
  extends AbstractListModel
  {
    private EditableDataContainer edc;
    private Vector v;

    public EDCListModel(EditableDataContainer edc) {
      super();
      this.edc = edc;
      v = new Vector(10, 5);
      update();
    }

    public int getSize() {
      return v.size();
    }

    public Object getElementAt(int i) {
      return v.elementAt(i);
    }

    public void update() {
      v.removeAllElements();
      for (Enumeration e = edc.getExistingElements(); e.hasMoreElements(); ) {
	v.addElement( (String)e.nextElement() );
      } // end for over e
      fireContentsChanged((Object)this, 0, getSize());
    }      
  }
}

