package NTLSystem.features;

import com.sun.java.swing.*;
import com.sun.java.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.*;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Vector;
import NTLSystem.NTLException;

/**
   A discrete feature-descriptor contains meta information that pertains
   to all instantiations of a particular discrete feature.
   <P>
   $Id: DiscreteFeatureDescriptor.java,v 1.19 1998/10/16 01:57:03 jsled Exp $
   <BR>Last editor: $Author: jsled $

   @author     Tajh Taylor [tajh@icsi.berkeley.edu]
   @version    $Revision: 1.19 $
 **/

public class DiscreteFeatureDescriptor
extends FeatureDescriptor
implements Serializable
{
    /**
     * serialization Stream Unique Identifier. Recompute [with
     * serialver] when this class changes.
     **/
    static final long serialVersionUID = 3695339423854179621L;

    // private instance variables

    /** A list of the value names; i.e. all possible values for the feature. */
    private String valueNames[];
    
    /** Whether or not this feature has ordered values. */
    private boolean ordered;

    // public instance methods

    // constructors

   /** A no-argument constructor for serialization. */
   public DiscreteFeatureDescriptor()
   {
     super();
     valueNames = new String[0];
     setDefaultIndex(new Integer(-1));
     ordered = false;
   }
  
   /** A simple determinate constructor.
       Assumes the default value is the empty string.
       @param newName        the new name of the feature
       @param newOrderedness whether or not this feature should be ordered;
                             probably doesn't matter in most cases */
   public DiscreteFeatureDescriptor(String newName, boolean newOrderedness)
   {
      super(newName);
      setOrderedness(newOrderedness);
      setDefaultIndex(new Integer(-1));
      valueNames = new String[0];
   }

   /** A detailed determinate constructor.
       @param     newName        the new name of this feature
       @param     newOrderedness whether or not this feature should be
                  ordered 
       @param     newValueNames  a list of the possible values
       @param     newDefault     the default value; should be in the
                  list 
       @exception NTLException If the newDefault value is not in the
                  newValueNames list.
       */
   public DiscreteFeatureDescriptor(String newName,
				    boolean newOrderedness,
				    String newValueNames[],
				    String newDefault)
       throws NTLException
  {
      super(newName);
      setOrderedness(newOrderedness);
      setValueNames(newValueNames);
      setDefaultValue(newDefault);
      Integer index = getValueIndex(newDefault);
      if (index.intValue() == -1) {
	throw new NTLException("Default value not found in value list.");
      } else {
	setDefaultIndex(index);
      }
   }

   /** A simple probabilistic constructor.
       @param newName        the new name of the feature
       @param newDefaultDist the new default discrete distribution
       @param newOrderedness whether or not this feature should be ordered;
                             probably doesn't matter in most cases */
   public DiscreteFeatureDescriptor(String newName,
				    DiscreteDist newDefaultDist,
				    boolean newOrderedness)
  {
    super(newName, newDefaultDist);
    setOrderedness(newOrderedness);
    valueNames = new String[newDefaultDist.getSize()];
    for(int i=0; i< valueNames.length; i++) {
      valueNames[i] = Integer.toString(i);
    } // end for over i
  }

   /** A detailed probabilistic constructor.
       @param     newName        the new name of this feature
       @param     newDefaultDist the default distribution; should be discrete
       @param     newOrderedness whether or not this feature should be ordered
       @param     newValueNames  a list of the possible values
       @exception NTLException if distribution size doesn't match
                  value list size. 
       */
   public DiscreteFeatureDescriptor(String newName,
				    DiscreteDist newDefaultDist, 
				    boolean newOrderedness,
				    String newValueNames[])
      throws NTLException
   {
      super(newName, newDefaultDist);
      setOrderedness(newOrderedness);
      setValueNames(newValueNames);
      if (newDefaultDist.getSize() != newValueNames.length) {
	throw new NTLException("discrete distribution size does not match value name array length");
      }
   }

   /** Override of Object.clone() that does the right thing. */
   public Object clone()
   {
     DiscreteFeatureDescriptor dfd =
       (DiscreteFeatureDescriptor)super.clone();
     dfd.valueNames = new String[valueNames.length];
     for (int i=0; i<valueNames.length; i++) {
       dfd.valueNames[i] = new String(valueNames[i]);
     }
     dfd.setOrderedness(getOrderedness());
     return dfd;
   }

   // implementation of abstract FeatureDescriptor methods

   /** Tell whether or not values for this feature are ordered.
       @return boolean flag indicating orderedness */
   public boolean isOrdered()
   {
      return ordered;
   }

   // class-specific methods

   /** Set a value name for a specific index.
       Does nothing if the index is invalid.
       @param index        the index at which to place this new value name
       @param newValueName the new value name to add */
   private void setValueName(Integer index, String newValueName)
   {
      if ( (index.intValue() >= 0) &&
	   (index.intValue() < valueNames.length) ) {
	 valueNames[index.intValue()] = newValueName;
      }
   }

   /** 
    * Set all the value names at once.
    * <P>
    * NOTE: This should be private, but it needs to be public for the
    * inner class DFDViewer.saveValues(), below.  If inner-class
    * inheritance gets fixed, this can be made private again.
    * <P>
    * <STRONG>This method should not be used, as it will be made
    * private at some point in the future.</STRONG>
    *
    * @param     newValueNames an array of strings; they become value
    *            names.
    * @exception NTLException if the array isn't fully populated.
    **/
   public void setValueNames(String newValueNames[])
      throws NTLException
   {
      int i;
      String tempValueNames[];

      tempValueNames = new String[newValueNames.length];
      for (i = 0; i < newValueNames.length; i++) {
	if (newValueNames[i] == null)
	  throw new NTLException("Value name array must be fully populated.");
	tempValueNames[i] = newValueNames[i];
      }
      valueNames = tempValueNames;
   }

   /** Get a value name at a specific index.
       Returns an empty string if the index is bad.
       @param     index The index of the desired value
       @return    The corresponding value string or null if one does
                  not exist.  * */
   public String getValueName(Integer index)
   {
      if ( (index.intValue() >= 0) &&
	   (index.intValue() < valueNames.length) ) {
	 return valueNames[index.intValue()];
      } else {
	 return new String("[ no value ]");
      }
   }

   /** Find the (first) index of a specific value string;
       return -1 if not found.
       @param searchValueName the value name to search for
       @return                the index of the value, or -1 if not found */
   public Integer getValueIndex(String searchValueName)
   {
      int i;
      int result = -1;
      boolean found;

      found = false;
      for (i=0; (i < valueNames.length) && (!found); i++) {
	if (valueNames[i].equals(searchValueName)) {
	  result = i;
	  found = true;
	}
      } // end for over i

      return new Integer(result);
   }

   /** Get all the value names at once.
       @return an array of strings containing the value names */
   public String[] getValueNames()
   {
      String retValueNames[];
      int i;

      retValueNames = new String[valueNames.length];
      for (i=0; i < valueNames.length; i++)
	 retValueNames[i] = valueNames[i];

      return retValueNames;
   }

   /** Report on how many value names there are.
       @return an integer count of value names */
   public Integer getValueCount()
   {
      return new Integer(valueNames.length);
   }

   /**
    * Set whether or not this feature has ordered values.
    * <P>
    * NOTE: This method should be private [or protected], but needs to
    * be public for DFDViewer.saveValues() [below].  This is a
    * workaround because inner-class inheritance is screwed.  If
    * inner-class inheritance gets fixed, this can be made private
    * again. 
    * <P>
    * <STRONG>Do not use this method, as it will be made private at
    * some point in the future, breaking your code.</STRONG>
    *
    * @param     newOrdered boolean flag indicating orderedness
    **/
   public void setOrderedness(boolean newOrdered)
   {
      ordered = newOrdered;
   }

   /** Report on whether or not this feature has ordered values.
       @return boolean flag indicating orderedness */
   public boolean getOrderedness()
   {
      return ordered;
   }

  /**
   * Set the default value (by index) for this feature. An invalid
   * index does nothing. 
   * <P>
   * NOTE: This method should be private, but it's necessary to make
   * it public for DFDViewer.saveValues(), because inner-class
   * inheritance is screwed.  If inner-class inheritance is fixed,
   * this can be made private again [once the appropriate changes are
   * made to the FDViewer subclasses].
   * <P>
   * <STRONG>Do not use this method, as it will be made private again
   * someday, and your code will break.</STRONG>
   *
   * @param     newDefaultIndex the index of the new default value
   **/
   protected void setDefaultIndex(Integer newDefaultIndex)
   {
      if ( (newDefaultIndex.intValue() >= 0) &&
	   (newDefaultIndex.intValue() < getValueCount().intValue()) ) {
	defaultValue = newDefaultIndex;
      }
   }
  
   /** Set the default value (by value name) for this feature.
       An invalid (= not found) value name does nothing.
       @param newDefaultValue the index of the new default value */
   protected void setDefaultValue(String newDefaultValue)
   {
      setDefaultIndex(getValueIndex(newDefaultValue));
   }

  /**
   * Gets the default index for this DiscreteFeatureDescriptor.
   *
   * @return    The default index for this DiscreteFeatureDescriptor.
   **/
  public Integer getDefaultIndex()
  {
    if (defaultValue instanceof Integer) {
      return (Integer)defaultValue;
    } else {
      return null;
    }
  }

  /**
   * @see FeatureDescriptor
   **/
  public Object valueFromText(String textVal)
       throws IllegalArgumentException
  {
    if (textVal.equals("[ no value ]")) {
      return new Integer(-1);
    } else {
      Integer toRet;
      toRet = getValueIndex(textVal);
      if (toRet.equals(new Integer(-1))) {
	throw new IllegalArgumentException("Unknown value " + textVal);
      } else {
	return toRet;
      }
    }
  }

  /**
   * @see FeatureDescriptor
   **/
  public String valueAsText(Object o)
  {
    if (isProbabilistic()) {
      return null;
    } else { 
      return getValueName((Integer)o);
    }
    
  }

   /** Test to see if this descriptor is identical to another. */
   public boolean equals(Object o)
   {
      boolean areEqual;
      int i;
      
      if (super.equals(o) == false)
	 return false;
      DiscreteFeatureDescriptor dfd = (DiscreteFeatureDescriptor)o;
      if (ordered != dfd.ordered)
	 return false;

      if (! getDefaultValue().equals(dfd.getDefaultValue())) {
	return false;   
      }

      // check the value names
      if (! getValueCount().equals(dfd.getValueCount())) {
	 return false;
      }
      areEqual = true;
      for (i = 0; (i < getValueCount().intValue()) && areEqual; i++) {
	areEqual &= valueNames[i].equals(dfd.getValueName(new Integer(i)));
      }
      return areEqual;
   }

   /** Compare two values (in the case of ordered value names)
       and tell whether value1 is less than value2.
       @param     value1  the first value name
       @param     value2  the second value name
       @return    boolean flag indicating that value1 < value2
       @exception NTLException for invalid value names or unordered
                  features. */
   public boolean lessThan(String value1, String value2)
      throws NTLException
   {
      int vi1, vi2;

      // check for error conditions
      if (!ordered)
	 throw new NTLException("Can't compare values for unordered features.");
      vi1 = getValueIndex(value1).intValue();
      vi2 = getValueIndex(value2).intValue();
      if ((vi1 == -1) || (vi2 == -1))
	 throw new NTLException("One or both value names are invalid.");

      // do the actual comparison
      return (vi1 < vi2);
   }

  /** Test whether the argument object contains a valid value for
       this particular feature.  Note that implementations should 
       report false for ProbDist object if the feature is determinate,
       and false for definite value object if the feature is
       probabilistic. 

       @param o any object to be tested.
       @return boolean flag indicating validity. */
   public boolean isValidValue(Object o)
   {
     if (isProbabilistic()) {
       if (o instanceof DiscreteDist) {
	 return true;
       } else {
	 return false;
       }
     } else {
       if (o instanceof Integer) {
	 return true;
       } else {
	 return false;
       }
     }
   }

   // implementation of Serializable methods
   // probably none is needed

   /** Read the descriptor in text form.
       @param input the character input source */
   public void readExternal(ObjectInput input)
      // throws IOException
   {
   }
  
   /** Write the descriptor in text form.
       @param output the character output source */
   public void writeExternal(ObjectOutput output)
      // throws IOException
   {
   }
  
  public String getDisplayableName()
  {
    return new String("Discrete Feature Descriptor");
  }

  protected JPanel getEditViewPanel(boolean edit)
  {
    viewer = new DFDViewer(this, edit);
    return viewer;
  }
} // end class DiscreteFeatureDescriptor

/* INNER CLASS for editing. */

class DFDViewer
    extends FDViewer
{
    /**
     * The default being-edited prob dist.
     **/
    private DiscreteDist defDD;
    /**
     * The possible discrete values.
     **/
    private Vector values;
    /**
     * The current default selection.
     **/
    private int curDefIdx;
    /**
     * The item name/probability mass/default selection table.
     **/
    private JTable dTable;
    /**
     * Ordered radiobutton.
     **/
    private JRadioButton ord;
    /**
     * Unordered radiobutton.
     **/
    private JRadioButton unord;
    /**
     * The "Up" ordering button
     **/
    private JButton up;
    /**
     * The "Down" ordering button
     **/
    private JButton down;
    /**
     * The new item addition text field.
     **/
    private JTextField newItemTF;
    /**
     * The DFDTableModel [see end of this inner class for definition]
     * used for editing/displaying all the data. 
     **/
    private DFDTableModel tableMod;

    public DFDViewer(FeatureDescriptor this_fd, boolean edit)
    {
	super(this_fd, edit);
    }

    public void createOptionalPanel()
    {
	DiscreteFeatureDescriptor dfd = (DiscreteFeatureDescriptor)fd;

	if (dfd.getDefaultIndex() != null) {
	    curDefIdx = dfd.getDefaultIndex().intValue();
	}
      
	if (dfd.isProbabilistic()) {
	    defDD = (DiscreteDist)dfd.getDefaultValue();
	    if (defDD != null) {
		// We really want a copy of the DiscreteDist we can play with.
		defDD = (DiscreteDist)defDD.clone();
	    }
	}
	if (defDD == null) {
	    // If one doesn't exist, create it.
	    defDD = new DiscreteDist( dfd.getValueNames().length );
	}

	// Get our own copy of the value names
	String vals[] = dfd.getValueNames();
	values = new Vector(vals.length, 1);
	// Copy the values into the Vector.
	for (int i=0; i<vals.length; i++) {
	    values.addElement(vals[i]);
	}

	JLabel l;
	JButton b;
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	optPanel = new JPanel(gbl);

	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Orderedness:");
	gbl.setConstraints(l, gbc);
	optPanel.add(l);
      
	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	JPanel ordUnordPanel = new JPanel();
	ButtonGroup obg = new ButtonGroup();
	ord = new JRadioButton("Ordered");
	ord.setActionCommand("ORD");
	ord.addActionListener(this);
	unord = new JRadioButton("Unordered");
	unord.setActionCommand("UNORD");
	unord.addActionListener(this);
	obg.add(ord);
	obg.add(unord);
	// Initially select the correct radiobutton
	if (dfd.isOrdered()) {
	    ord.setSelected(true);
	    unord.setSelected(false);
	} else {
	    ord.setSelected(false);
	    unord.setSelected(true);
	}
	// Disable if we're in 'viewing' mode.
	if (viewing) {
	    ord.setEnabled(false);
	    unord.setEnabled(false);
	}
	ordUnordPanel.add(ord);
	ordUnordPanel.add(unord);
	gbl.setConstraints(ordUnordPanel, gbc);
	optPanel.add(ordUnordPanel);

	// Create the table model...
	tableMod = new DFDTableModel();
	// Create the table
	dTable = new JTable(tableMod);
	// Make the table single-selection only
	// [required for "Up/Down" ordering functionality].
	dTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	// Make the table have a good default size.
	dTable.setPreferredScrollableViewportSize(new Dimension(300, 100));
	// Put it in a JScrollPane
	JScrollPane jsp = new JScrollPane(dTable);

	gbc.weightx = gbc.weighty = 1.0;
	gbc.fill = GridBagConstraints.BOTH;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbc.anchor = GridBagConstraints.CENTER;
	gbl.setConstraints(jsp, gbc);
	optPanel.add(jsp);

	gbc.weightx = gbc.weighty = 0.0;

	if (!viewing) {
	    JPanel bpanel = new JPanel();

	    up = new JButton("Up");
	    up.setActionCommand("UP");
	    up.addActionListener(this);
	    bpanel.add(up);

	    down = new JButton("Down");
	    down.setActionCommand("DOWN");
	    down.addActionListener(this);
	    bpanel.add(down);

	    // Set the initial state of the up/down buttons correctly [ie:
	    // if we're unordered, disable the buttons].
	    if (!dfd.isOrdered()) {
		up.setEnabled(false);
		down.setEnabled(false);
	    }
	
	    b = new JButton("Remove");
	    b.setActionCommand("REMOVE");
	    b.addActionListener(this);
	    bpanel.add(b);
	
	    b = new JButton("Normalize");
	    b.setActionCommand("NORM");
	    b.addActionListener(this);
	    bpanel.add(b);

	    gbc.fill = GridBagConstraints.NONE;
	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    gbc.anchor = GridBagConstraints.CENTER;
	    gbl.setConstraints(bpanel, gbc);
	    optPanel.add(bpanel);

	    gbc.fill = GridBagConstraints.NONE;
	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    JPanel apanel = new JPanel();

	    newItemTF = new JTextField( 25 );
	    gbl.setConstraints(newItemTF, gbc);
	    newItemTF.addActionListener(this);
	    apanel.add(newItemTF);
	
	    b = new JButton("Add");
	    b.setActionCommand("ADD");
	    b.addActionListener(this);
	    apanel.add(b);
	
	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    gbl.setConstraints(apanel, gbc);
	    optPanel.add(apanel);
	}
    }

    public void actionPerformed(ActionEvent ae)
    {
	String acmd = ae.getActionCommand();
	// Treat a <RET> on the text field as an "ADD", for ease-of-use.
	if ( (ae.getSource() == newItemTF) ||
	     acmd.equals("ADD")) {
	    // Get the value from the text field
	    String toAdd = newItemTF.getText();
	    // Add it to the values
	    values.addElement( (Object)toAdd );
	    // Add it to the distribution
	    defDD.add();
	    // Update the table
	    tableMod.rowsInserted(values.size()-1, values.size()-1);
	    // Remove the value from the textfield.
	    newItemTF.setText("");
	} else if (acmd.equals("REMOVE")) {
	    // Get the selection index
	    int selIdx = dTable.getSelectedRow();
	    // Remove it from the values
	    values.removeElementAt(selIdx);
	    // Remove it from the distribution
	    defDD.remove(selIdx);
	    // Update the table
	    tableMod.rowsDeleted(selIdx, selIdx);
	} else if (acmd.equals("UP")) {
	    // Get the selection index
	    int selIdx = dTable.getSelectedRow();
	    if (selIdx > 0) {
		//  swap it with index-1 in values
		Object tmp = values.elementAt(selIdx-1);
		values.setElementAt(values.elementAt(selIdx), selIdx-1);
		values.setElementAt(tmp, selIdx);
		//  swap it in the dist.
		defDD.swap(selIdx, selIdx-1);
		// Select the new row
		dTable.setRowSelectionInterval(selIdx-1, selIdx-1);
	    }
	    // Table update
	    tableMod.rowsUpdated(selIdx-1, selIdx);
	} else if (acmd.equals("DOWN")) {
	    // Get the selection index
	    int selIdx = dTable.getSelectedRow();
	    if (selIdx < (values.size()-1)) {
		//  swap it with index+1 in values
		Object tmp = values.elementAt(selIdx+1);
		values.setElementAt(values.elementAt(selIdx), selIdx+1);
		values.setElementAt(tmp, selIdx);
		//  swap it in the dist.
		defDD.swap(selIdx, selIdx+1);
		// Select the new row
		dTable.setRowSelectionInterval(selIdx+1, selIdx+1);
	    }
	    // Table update
	    tableMod.rowsUpdated(selIdx, selIdx+1);
	} else if (acmd.equals("NORM")) {
	    // If editing, stop table editing.
	    if (dTable.getCellEditor() != null) {
		dTable.getCellEditor().stopCellEditing();
	    }
	    // Normalize the values.
	    defDD.normalize();
	    // table update
	    tableMod.dataUpdated();
	} else if (acmd.equals("ORD")) {
	    // enable the up/down buttons.
	    up.setEnabled(true);
	    down.setEnabled(true);
	} else if (acmd.equals("UNORD")) {
	    // disable the up/down buttons.
	    up.setEnabled(false);
	    down.setEnabled(false);
	} else {
	    super.actionPerformed(ae);
	}
    }

    public void createProbPanel()
    {
	; // noop; everything is handled in the optional panel's table
    }

    public void createDetPanel()
    {
	; // noop; everything is handled in the optional panel's table
    }

    public String getFDTypeName()
    {
	return new String("Discrete Feature Descriptor");
    }

    public void canSaveValues(Vector reasons)
    {
	// Yes, we can save our values.  There's always _some_ default
	// selection, and the probability masses in the table are
	// checked on input.
	super.canSaveValues(reasons);
    }

    public void saveValues()
	throws CannotSaveValuesException
    {
	if (viewing) { return; }
	super.saveValues();
	// Save our values:
	// ordered, values, if prob (probvalues), if det (detvalue)
	DiscreteFeatureDescriptor dfd = (DiscreteFeatureDescriptor)fd;
	dfd.setOrderedness(ord.isSelected());
	String newVals[] = new String[values.size()];
	values.copyInto(newVals);
	try {
	    dfd.setValueNames(newVals);
	} catch (NTLException ntle) {
	    throw new CannotSaveValuesException(ntle.getMessage());
	} // end try-catch
	if (prob.isSelected()) {
	    dfd.defaultValue = defDD;
	    dfd.setProbabilistic(true);
	} else {
	    dfd.setDefaultIndex( new Integer(curDefIdx) );
	    dfd.setProbabilistic(false);
	}
    }

    class DFDTableModel
	extends AbstractTableModel {

	private NumberFormat nf;

	public DFDTableModel() {
	    nf = NumberFormat.getNumberInstance();
	    nf.setMaximumFractionDigits(5);
	}

	public int getRowCount() {
	    return values.size();
	}
      
	public int getColumnCount() {
	    if (viewing) {
		return 2;
	    } else {
		return 3;
	    }
	}

	public Class getColumnClass(int columnIndex) {
	    Class toRet = null;
	    int c = fixColumnIndex(columnIndex);
	    switch (c) {
	    case 0:
		toRet = java.lang.String.class;
		break;
	    case 1:
		toRet = java.lang.Double.class;
		break;
	    case 2:
		toRet = java.lang.Boolean.class;
		break;
	    }
	    return toRet;
	}

	// 'Fix' the column number.  Columns are numbered logically
	// as such:
	// 0: Name
	// 1: PMass [DiscreteDist default value]
	// 2: default selection [Deterministic default value]
	//
	// However, if we're in viewing mode, there are only two
	// columns... to keep switch statements simple, we adjust the
	// column value to conform to our logical column numbering.
	// 
	// Note that this is only a problem for column 1... column 0
	// is always the name, and column 2 is unambiguous [always
	// the default selection].
	private int fixColumnIndex(int ci) {
	    if ( (ci==1) &&
		 (viewing && !(((DiscreteFeatureDescriptor)fd).isProbabilistic())) ) {
		return ci+1;
	    } else {
		return ci;
	    }
	}

	public String getColumnName(int columnIndex) {
	    String toRet = null;
	    int c = fixColumnIndex(columnIndex);
	    switch (c) {
	    case 0:
		toRet = new String("Element name");
		break;
	    case 1:
		toRet = new String("Probability mass");
		break;
	    case 2:
		toRet = new String("Default selection?");
		break;
	    }
	    return toRet;
	}

	public Object getValueAt(int row, int column) {
	    Object toRet = null;
	    int c = fixColumnIndex(column);
	    switch (c) {
	    case 0:
		toRet = values.elementAt(row);
		break;
	    case 1:
		// Format it as a string with 5 sigdigits.
		String strVal = nf.format(defDD.getProb(row));
		// Convert that string to a Double
		toRet = new Double(strVal);
		break;
	    case 2:
		toRet = new Boolean( (curDefIdx == row) );
		break;
	    }
	    return toRet;
	}

	public void setValueAt(Object aValue,
			       int rowIndex, int columnIndex) {
	    int c = fixColumnIndex(columnIndex);
	    switch (c) {
	    case 0:
		values.setElementAt(aValue, rowIndex);
		break;
	    case 1:
		try {
		    // Get a Number from the String
		    Number n = nf.parse((String)aValue);
		    // Set the double value of the Number
		    defDD.setProb(rowIndex, n.doubleValue());
		} catch (ParseException pe) {
		    // Alert the user that the value is ... invalid.
		    JOptionPane.showMessageDialog(null, "Field cannot have the value " +
						  (String)aValue + ".");
		} // end try-catch
		break;
	    case 2:
		int oldIdx = curDefIdx;
		curDefIdx = rowIndex;
		fireTableRowsUpdated(oldIdx, oldIdx);
		fireTableRowsUpdated(rowIndex, rowIndex);
		break;
	    }
	}

	public boolean isCellEditable(int rowIndex, int columnIndex) {
	    // If we're viewing, nothing is editable.
	    if (viewing) {
		return false;
	    }

	    return true;
	}
      
	public void rowsInserted(int rowA, int rowB) {
	    fireTableRowsInserted(rowA, rowB);
	}
	public void rowsDeleted(int rowA, int rowB) {
	    fireTableRowsDeleted(rowA, rowB);
	}
	public void rowsUpdated(int rowA, int rowB) {
	    fireTableRowsUpdated(rowA, rowB);
	}
	public void dataUpdated() {
	    fireTableDataChanged();
	}
    } // end inner-inner-class DFDTableModel
} // end inner-class DFDViewer
