package NTLSystem.features;

import java.io.*;
import java.lang.*;

/**
 * Feature is the class that corresponds to one slot in an f-struct.
 * to all instantiations of a particular feature.
 * <P>
 * Features can have a null value, which indicates that there is
 * currently [suprise suprise] no value for this Feature.  However,
 * newly-created Features have the default value from their
 * FeatureDescriptor.
 * <P>
 * $Id: Feature.java,v 1.17 1998/10/16 01:57:04 jsled Exp $
 * <BR>
 * Last editor: $Author: jsled $
 *
 * @author     Tajh Taylor [tajh@icsi.berkeley.edu]
 * @version    $Revision: 1.17 $
 **/
public class Feature
implements Serializable, Cloneable
{
   // static class fields

   /** A static class field; turns on/off value security checking */
   public static boolean checkValues;

   // private instance fields

   /** A pointer to the descriptor that this was spawned from */
    protected FeatureDescriptor descriptor;

   /** Holds either a specific value or a probability distribution
       for the feature.
       All type-checking is done by the feature descriptor. */
    protected Object value;

    /**
     * The name of this Feature.  If 'null', then the name is the same
     * as the FeatureDescriptor.
     **/
    protected String name;

   /** A pointer to the containing Fstruct;
       allows belief maintenance to be performed when necessary. */
    protected Fstruct container; 

   // static initializer
   static
   {
      checkValues = true; // default
   }

   // public instance methods

   // instance constructors

   /** A plain no-argument constructor for serialization.
       A feature constructed this way won't be very usable. */
   public Feature()
   {
      container = null;
      value = null;
      descriptor = null;
      name = null;
   }

  /**
   * Constructs a Feature from the FeatureDescriptor.  The default
   * value of the FeatureDescriptor is cloned and used for the value
   * of the Feature.
   *
   * @param     newDescriptor The FeatureDescriptor from which to base
   *            this Feature.
   **/
  public Feature(FeatureDescriptor newDescriptor)
  {
    container = null;
    name = null;
    descriptor = newDescriptor;
    if (descriptor.isProbabilistic()) {
      value = ((ProbDist)newDescriptor.getDefaultValue()).clone();
    } else {
      Object defVal = newDescriptor.getDefaultValue();
      if (defVal instanceof Integer) {
	value = new Integer( ((Integer)defVal).intValue() );
      } else if (defVal instanceof Double) {
	value = new Double( ((Double)defVal).doubleValue() );
      } else if (defVal instanceof String) {
	value = new String( (String)defVal );
      }
    }
  }

   /** The regular constructor.
       @param     newDescriptor the new descriptor for the feature
       @param     newValue the new value for the feature
       @exception IllegalArgumentException If the newValue is not
                  valid for the Feature, given the newDescriptor.
		  **/
   public Feature(FeatureDescriptor newDescriptor, Object newValue)
      throws IllegalArgumentException
   {
      container = null;
      name = null;
      descriptor = newDescriptor;

      // see if we should be checking values
      if (checkValues)
      {
         if (!descriptor.isValidValue(newValue))
	   throw new IllegalArgumentException("Invalid newValue given for Feature");
      }
      value = newValue;
   }

   //===================================================================
   // access methods (get/set routines)
   //
   // the current suggested method for belief updates:
   // the setValue method should only modify the probability distribution
   // directly if there is no containing Fstruct; otherwise, it should
   // post a distribution belief change to the Fstruct, and then the
   // Fstruct should call commitProbDist to actually "burn in" the new
   // distribution.  This prevents Features from modifying themselves
   // in a manner inconsistent with the belief state.
   // note that when getProbDist is called, the parent Fstruct should
   // call its commitBeliefUpdates() method to ensure that belief
   // consistency is maintained.

   //TODO : commented out until further understood
   /*protected void commitProbDist(ProbDist newDist)
      throws NTLException
   {
      if (checkValues)
      {
         if (!descriptor.isValidValue(newDist))
	    throw new NTLException("invalid value given for feature(2)");
      }
      value = (Object)newDist;
   }*/
  
   /**
    * Looks in the feature dictionary to get a matching descriptor and then
    * changes the feature to match; also blanks value if successful.
    * <P>
    * FIXME: Check with tajh before removing/rewriting...
    *
    * @param     newName  the name of the desired new feature
    * @param     fDict    the feature dictionary to look in
    * @exception NoSuchFeatureException If the given Feature name
    *            cannot be found in the FeatureDictionary.
    **/
   public void setName(String newName, FeatureDictionary fDict)
      throws NoSuchFeatureException
   {
      // find a matching descriptor first
     if (!fDict.isDescriptor(newName)) {
       throw new NoSuchFeatureException("Feature \"" + newName +
					"\" not found in dictionary.");
     }
     descriptor = fDict.getDescriptor(newName);
     value = null;
   }


   /** Get the name from the feature descriptor.
       @return return a string containing the feature name */
   public String getName()
   {
       if (name != null) {
	   return name;
       } else {
	   return descriptor.getName();
       }
   }

    /**
     * Sets the name of the Feature.  If the name is set to 'null',
     * then the effective name of the Feature [as returned by
     * 'getName()'] is the name of the FeatureDescriptor.
     * 
     * @param     newName The new name of the Feature, or null if the
     *            name should be the same as the FeatureDescriptor.
     **/
    public void setName(String newName)
    {
	name = newName;
    }

   /** Changes the feature descriptor to match.
       @param     newDescriptor the new feature descriptor
       @exception IllegalArgumentException If the newValue is not
                  allowed by the newDescriptor. **/
   public void setDescriptor(FeatureDescriptor newDescriptor)
      throws IllegalArgumentException
   {
      // see if we should be checking for valid values
      if (checkValues)
      {
	 if (!newDescriptor.isValidValue(value))
	    throw new IllegalArgumentException("Existing value invalid for " +
					       "newDescriptor; descriptor unchanged.");
      }
      descriptor = newDescriptor;
   }

   /** Changes the descriptor and the value at the same time.
       @param     newDescriptor the new feature descriptor
       @param     newValue      the new value
       @exception IllegalArgumentException If the newValue is invalid
                  for the newDescriptor.
		  **/
   public void setDescriptor(FeatureDescriptor newDescriptor,
			     Object newValue)
       throws IllegalArgumentException
  {
    // see if we should be checking for valid values
    if (checkValues) {
      if (!newDescriptor.isValidValue(newValue)) {
	throw new IllegalArgumentException("newValue invalid for newDescriptor; " +
					   "feature unchanged.");
      }
    }
  }
  
   /** Returns the feature descriptor being used.
       @return returns the feature descriptor */
   public FeatureDescriptor getDescriptor()
   {
      return descriptor;
   }

   // value access methods

  /**
   * Sets the value for this Feature as a String.  The
   * FeatureDescriptor is responsible for converting the String into
   * the appropriate object for this Feature.
   *
   * @param     newStringValue The new value as a String.
   * @exception IllegalArgumentException If the String value cannot be
   *            parsed or converted into the appropriate object.
   **/
  public void setValueAsText(String newStringValue)
       throws IllegalArgumentException
  {
    Object newVal = descriptor.valueFromText(newStringValue);
    setValue(newVal);
  }

  /**
   * Gets the value for this Feature [if deterministic] as a String,
   * by using the FeatureDescriptor.
   *
   * @return    The value for this Feature as a String.  If
   *            Probabilistic, returns null.
   **/
  public String valueAsText()
  {
    if (getDescriptor().isProbabilistic()) {
      return null;
    } else {
      if (getValue() == null) {
	return new String("[ no value ]");
      } else {
	return getDescriptor().valueAsText( getValue() );
      }
    }
  }

   /** Set the value for this feature.
       Throws an exception if the value is invalid.
       @param     newValue the desired new value for this feature
       @exception IllegalArgumentException If the newValue is not
                  valid with the current feature descriptor.
		  **/
   public void setValue(Object newValue)
      throws IllegalArgumentException
   {
     if (newValue == null) {
       // If it's null, just set it, and don't bother to check with
       // the descriptor for validity.
       value = null;
       return;
     } 

      // make sure we're dealing with a valid value
     if (!getDescriptor().isValidValue(newValue)) {
	throw new IllegalArgumentException("Invalid newValue given for feature."); 
     }

      // check if there's a containing fstruct that can handle belief
      // updating
      if ((newValue instanceof ProbDist) && (container != null))
      {
	 // TODO
	 // let the container handle this
	 // container.addBeliefUpdate(this, newValue);
	 // for now, just do it the old fashioned independent way
	 // this should be changed later
	 value = newValue;
      } else {
	// just do it ourselves
	value = newValue;
      }
   }

   /** Get the value for this feature.
       Must be casted to the proper type.  May be null.
       @return returns the current value */
   public Object getValue()
   {
      return value;
   }

  /**
   * Sets the containing Fstruct for this Feature.
   *
   * @param     fs The Fstruct which contains this feature.
   **/
  public void setContainer(Fstruct fs)
  {
    container = fs;
  }

  /**
   * Gets the containing Fstruct for this Feature.
   *
   * @return    The Fstruct which contains this Feature.
   **/
  public Fstruct getContainer()
  {
    return container;
  }

   /** Reports on whether or not this feature is inside a containing
       f-struct.
       @return    a boolean flag indicating container membership */
   public boolean isContainer()
   {
      return !(container == null);
   }
  
   // Utility

  /**
   * Redefiniton of Object.clone()
   **/
  public Object clone()
  {
    try {
	Feature f = (Feature)super.clone();
	f.descriptor = descriptor;
	if (value instanceof ProbDist) {
	    f.value = ((ProbDist)value).clone();
	} else if (value instanceof Integer) {
	    f.value = new Integer(((Integer)value).intValue());
	} else if (value instanceof Double) {
	    f.value = new Double(((Double)value).doubleValue());
	} else if (value instanceof String) {
	    f.value = new String(((String)value));
	} 
	f.container = this.container;
	if (this.name != null) {
	    f.name = new String( this.name );
	}
	return f;
    } catch (CloneNotSupportedException cnse) {
	// Shouldn't happen as all types of data which a Feature can
	// contain are cloneable.
	throw new InternalError(cnse.getMessage());
    } // end try-catch
  }

   /** Check this feature against another object for equality.
       @oaram o object to be tested
       @return boolean flag indicating equality */
   public boolean equals(Object o)
   {
      // first, see if o is of the proper class
     if ( !(getClass().isInstance(o)) ) {
       return false;
     }

     Feature otherFeature = (Feature)o;
      
      // now check the descriptors
     if (!descriptor.equals(otherFeature.getDescriptor()))
       return false;
	
      // now check the values
     if (!value.equals(otherFeature.getValue())) {
       return false;
     }

      // everything checked out; return true
     return true;
   }
  
  /**
   * Redefinition of Object.toString()
   **/
  public String toString()
  {
    StringBuffer toRet = new StringBuffer(100);
    toRet.append("Feature { Name: ");
    toRet.append(getName());
    toRet.append("; value: ");
    if (getValue() == null) {
      toRet.append("[ no value ]");
    } else {
      toRet.append( getValue().toString() );
    }
    toRet.append(" }");
    return toRet.toString();
  }

   /**
    * Peakedness of this feature. ** FIXME
    * A heuristic measure of the peakedness of the distribution.
    * Depends on distribution; haven't handled for continuous yet, Gaussian. FIXME!!
    * Return infinity for a deterministic value or distribution-dependent result.
    **/
   public double peakedness() { 
     if (value instanceof DiscreteDist) 
       {
	 return 0; // temporary
	 //	 return value.peakedness(); // need to define!
       }
     else return Double.POSITIVE_INFINITY;
   }

  /**
   * Return the highest-probability feature value for this distribution.
   * FIXME!!!! Before returned string value name, won't work now...
   * need to update...
   **/
  // 
  public Object mode() {
//     String leadingVal = null;
//     double leadingCount = -1;
//     Enumeration valEnum = counts.keys();
//     while (valEnum.hasMoreElements()) {
//       String val = (String)valEnum.nextElement();
//       double count = getCount(val);
//       if (count > leadingCount) {
// 	leadingCount = count;
// 	leadingVal = val;
//       }
//     }
//     return leadingVal;
    return null;
  }


  // Externalizable
  // FIXME: When externalizable is impl...
  //public void readExternal(ObjectInput input) {}
  //public void writeExternal(ObjectOutput output) {}
}
