package NTLSystem.features;

import java.io.*;
import java.util.*;
import NTLSystem.*;

/**
 * A FeatureDictionary contains 0..n FeatureDescriptors representing
 * possible Features in the system.
 * <P>
 * $Id: FeatureDictionary.java,v 1.8 1998/10/16 01:57:06 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       EditableDataDictionary
 * @see       FeatureDescriptor
 * @version   $Revision: 1.8 $
 **/
public class FeatureDictionary
    extends EditableDataDictionary
    implements Serializable
{

    // serialization stream unique identifier
    // recompute [with serialver] when this class changes.
    static final long serialVersionUID = -5126178739094602841L;

  /* CONSTRUCTORS */

  /**
   * Basic constructor for a FeatureDictionary.
   **/
  public FeatureDictionary()
  {
    super();
  }

  /**
   * Puts a descriptor in the FeatureDictionary.   If this replaces an
   * entry in the FeatureDictionary, so be it.
   *
   * @param     fd The FeatureDescriptor to put in the
   *            FeatureDictionary.
   **/
  public void putDescriptor(FeatureDescriptor fd)
  {
    set(fd.getKey(), (EditableDataElement)fd);
  }

  /**
   * Gets the Feature Descriptor with the given name from the
   * FeatureDictionary. Returns null if the object does not exist in
   * the FeatureDictionary.
   *
   * @param     searchName The name of the FeatureDescriptor to search
   *            for.
   * @return    The FeatureDescriptor with the given name, or null if
   *            it does not exist.
   **/
  public FeatureDescriptor getDescriptor(String searchName)
  {
    return (FeatureDescriptor)get(searchName);
  }

  /**
   * Test to see if a particular named FeatureDescriptor is in the
   * FeatureDictionary.
   *
   * @param     searchName The name of the FeatureDescriptor to test
   *            for.
   * @return    True if the FeatureDescriptor is in the
   *            FeatureDictionary, false if not.
   **/
  public boolean isDescriptor(String searchName)
  {
    return containsKey(searchName);
  }

  /**
   * Test to see if the particular FeatureDescriptor object is in the
   * FeatureDictionary.
   *
   * @param     otherFD The FeatureDescriptor to check for.
   * @return    true If the FeatureDescriptor is in the
   *            FeatureDictionary, false if not.
   **/
  public boolean isDescriptor(FeatureDescriptor otherFD)
  {
    return contains(otherFD);
  }
    
  /**
   * Returns an Enumeration of all the FeatureDescriptors in the
   * FeatureDictionary. 
   *
   * @return    An Enumeration of all the FeatureDescriptors.
   **/
  public Enumeration getAllDescriptors()
  {
    return dict.elements();
  }

  /**
   * Returns the count of the number of FeatureDescriptors in the
   * FeatureDictionary. 
   *
   * @return    The number of FeatureDescriptors in the
   *            FeatureDictionary. 
   **/
  public int countDescriptors()
  {
    return dict.size();
  }

  // generate a new f-struct containing all the features in the dictionary,
  // and give them all the default values or distributions
  //
  // How useful is this?
  public Fstruct generateFstruct()
  {
    return null;
  }

  // generate a new Fstruct containing the given subset of Features
  // in the dictionary.
  public Fstruct generateFstruct(FeatureDescriptor[] fds)
  {
    return null;
  }

  /**
   * Returns a single Feature from the FeatureDictionary.  This
   * feature is constructed from the given FeatureDescriptor name and
   * that FeatureDescriptor's default value.
   *
   * @param     fdName The name of the FeatureDescriptor to base the
   *            Feature on.
   * @return    A new Feature constructed from the given
   *            FeatureDescriptor name, or null if no such named
   *            FeatureDescriptor exists.
   **/
  public Feature getFeature(String fdName)
  {
    FeatureDescriptor fd = getDescriptor(fdName);
    if (fd == null) {
      return null;
    } 

    // Create a new Feature from the FeatureDescriptor
    return new Feature(fd);
  }


  /*  Implement abstract methods in EditableDataDictionary  */

  /**
   * Returns the collection of allowable types which can be stored in
   * the FeatureDictionary.  This is the collection of Classes which
   * subclass FeatureDescriptor.
   *
   * @return    An array of Classes which subclass FeatureDescriptor.
   **/
  protected Class[] getAllowableTypes()
  {
    try {
      Class toRet[] =
      { Class.forName("NTLSystem.features.ContinuousFeatureDescriptor"), 
	Class.forName("NTLSystem.features.IntegralFeatureDescriptor"),
	Class.forName("NTLSystem.features.DiscreteFeatureDescriptor"),
	Class.forName("NTLSystem.features.StringFeatureDescriptor") };
      return toRet;
    } catch (ClassNotFoundException cnfe) {
      // Shouldn't happen
      throw new InternalError();
    } // end try-catch
  }

    /**
     * A static method to return a new FeatureDictionary by
     * deserializing it from the specified file name.
     *
     * @param     file The File object from which to load the
     *            FeatureDictionary.
     * @return    The FeatureDictionary loaded from the given file.
     * @exception FileNotFoundException If the given file cannot be
     *            found. 
     * @exception ClassNotFoundException If a class which has been
     *            serialized in the file cannot be found.
     * @exception IOException If there is an IOException in the
     *            underlying stream while reading the data.
     **/
    public static FeatureDictionary loadDict(File f)
	throws FileNotFoundException, ClassNotFoundException,
	IOException
    {
	ObjectInputStream ois =
	    new ObjectInputStream(new BufferedInputStream(new FileInputStream( f )));
	FeatureDictionary featureDict =
	    (FeatureDictionary)ois.readObject();
	ois.close();
	return featureDict;
    }

    /**
     * Serializes the FeatureDictionary to the given File.  If the File
     * exists, it will be over written.
     *
     * @param     toSaveTo The File object pointing to the File where
     *            this FeatureDictionary should be saved.
     * @exception IOException If something goes wrong while saving.
     **/
    public void saveDict( File toSaveTo )
	throws IOException
    {
	ObjectOutputStream oos =
	    new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream( toSaveTo )));
	oos.writeObject( this );
	oos.close();
    }


  /**
   * Testing code.
   **/
  public static void main(String args[])
  {
    // Get the properties.
    String propFN = new String("properties/default.properties");
    Properties ourProps = new Properties(System.getProperties());
    try {
      FileInputStream propsIn = new FileInputStream(propFN);
      ourProps.load(propsIn);
      System.setProperties(ourProps);
    } catch (SecurityException se) {
      // Shouldn't happen as an app.
      throw new InternalError(se.getMessage());
    } catch (FileNotFoundException fnfe) {
      System.out.println("Could not find NTLSystem properties file " + propFN);
    } catch (IOException ioe) {
      System.out.println("Could not load default properties file.");
    } // end try-catch

    String dictFN = ourProps.getProperty("ntlsys.dataDir") +
      "/" + ourProps.getProperty("ntlsys.dictDir") +
      "/";
    if (args.length == 0) {
	dictFN += ourProps.getProperty("ntlsys.featureDictName");
    } else {
	dictFN += args[0];
    }
    FeatureDictionary fd = null;

    try {
      FileInputStream fis =
	new FileInputStream(dictFN);
      ObjectInputStream ois = new ObjectInputStream(fis);
      fd = (FeatureDictionary)ois.readObject();
    } catch (IOException ioe) {
      ioe.printStackTrace();
      System.out.println(ioe.getMessage());
    } catch (ClassNotFoundException cnfe) {
      // Won't happen
    }
    
    if (fd == null) {
      System.out.println("Feature Dictionary _not_ restored.");
      fd = new FeatureDictionary();
    }
    fd.Edit(true);

    try {
      FileOutputStream fos = new FileOutputStream(dictFN);
      ObjectOutputStream oos = new ObjectOutputStream(fos);
      oos.writeObject(fd);
      oos.flush();
      fos.close();
    } catch (IOException ioe) {
      ioe.printStackTrace();
      System.out.println(ioe.getMessage());
    }

    System.exit(0); 


 }

}
