package NTLSystem.features;

import com.sun.java.swing.*;
import com.sun.java.swing.border.*;
import com.sun.java.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Vector;
import NTLSystem.support.WindowWaiter;

/**
 * The PropertyEditor for Features contained in a
 * PropertyCollectionSheetPanel.  This class controls editing of the
 * Feature, including all possible values which the Feature may
 * contain.  The FeatureDescriptor which the Feature contains is very
 * useful here to figure out how to edit the Feature.
 * <P>
 * A lot of this editing code is duplicated from similar code in
 * FeatureDescriptor editing... specifically, the JTable-based
 * Feature editing for DiscreteFeatureDescriptor-based Features.
 * There should be a more abstract way of editing such values.  As
 * well, the Distribution-changing code is duplicated [somewhat] from
 * FeatureDescriptor editing.  It would be nice if this was merged
 * somehow.
 * <P>
 * $Id: FeatureEditor.java,v 1.8 1998/10/16 01:57:07 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Feature
 * @see       FeatureDescriptor
 * @see       NTLSystem.support.PropertyCollectionSheetPanel
 * @version   $Revision: 1.8 $
 **/
public class FeatureEditor
extends PropertyEditorSupport
{

    /* ATTRIBUTES */
    
    /** The Feature being edited. **/
    private Feature value;

    /** The Fstruct containing that Feature. **/
    private Fstruct fst;

    /** The FeatureDescriptor from that feature. **/
    private FeatureDescriptor fd;

    /** A Panel used for the custom editor. **/
    private JPanel jp;

    /** A Component used for the custom editor. **/
    private Component eipc;


    
    public void setValue(Object newValue)
    {
	value = (Feature)newValue;
	fd = value.getDescriptor();
	fst = value.getContainer();
    }

    public Object getValue()
    {
	return value;
    }

  /**
   * This returns a String-representation of the value of the Feature
   * for determinisitic Features.  For deterministic Discrete
   * Features, this is the default value name.
   *
   * @return     A String representation of the Feature's value.
   **/
  public String getAsText()
  {
    // If not probabilistic
    if ( !fd.isProbabilistic() ) {
      return value.valueAsText();
    } else {
      return null;
    }
  }

  /**
   * This sets the Feature's value as a String.  The empty String is
   * used to represent a null Feature value.  The FeatureDescriptor is
   * used to parse the String value into a suitable Object for setting
   * the Feature with.
   *
   * @param     newStringVal The new value for this Feature to take
   *            on, as a String.
   * @exception IllegalArgumentException If the newStringVal cannot be
   *            parsed into a suitable value by the FeatureDescriptor.
   **/
  public void setAsText(String newStringVal)
       throws IllegalArgumentException
  {
    // Only if not probabilistic
    if (!fd.isProbabilistic()) {
      if (newStringVal.equals("")) {
	value.setValue( null );
      } else {
	value.setValue( fd.valueFromText(newStringVal) );
      } 
    }
  }

  /**
   * Returns the possible values for a deterministic Discrete
   * Feature.  In addition, the String "[ no value ]" is added to the
   * returned array to indicate a null Feature value.
   *
   * @return    An array of Strings containing the possible values for
   *            a deterministic Discrete Feature(Descriptor).
   **/
  public String[] getTags()
  {
    if (!fd.isProbabilistic() &&
	(fd instanceof DiscreteFeatureDescriptor)) {
      // Return the possible value names
      int count = ((DiscreteFeatureDescriptor)fd).getValueCount().intValue();
      String toRet[] = new String[count+1];
      System.arraycopy(((DiscreteFeatureDescriptor)fd).getValueNames(), 0,
		       toRet, 0, count);
      // Set the last index to "[ no value ]" for an undefined index.
      toRet[count] = "[ no value ]";
      return toRet;
    } else {
      return null;
    }
  }

  /**
   * Returns a Component to be embedded in the
   * PropertyCollectionSheetPanel as a custom editor for the Feature.
   * This is the case for all probabilistic Features.
   * <P>
   * In the case of Integral and Continuous Features, the ProbDist's
   * <TT>getEditInplacePanel()</TT> method is used to get the editor.
   * For Discrete Features, small editor is returned which displays
   * the possible values, sorted by decreasing probability.  In this
   * case, the ability for an editor to be brought up is also given,
   * which brings up a window with a JTable which allows the editing
   * of the Pmasses.
   * <P>
   * In any case, the returned Component can be right-clicked on to
   * bring up a PopupMenu which contains the ability to:
   * <UL>
   * <LI>Set the Feature to have no distribution [null value].
   * <LI>Set the Feature to have the original defalut distribution.
   * <LI>Change the Feature to have a new distibution [from the set of
   *     allowable types].
   * </UL>
   *
   * @return    A Component which provides the ability to edit the
   *            Feature's values.
   **/
  public Component getCustomEditor()
  {
      JComponent toRet = null;

      if (fd.isProbabilistic()) {
	  GridBagLayout gbl = new GridBagLayout();
	  jp = new JPanel(gbl);
	  // Return a component with the type of FD being edited.
	  if (value.getValue() != null) {
	      jp.setToolTipText(((ProbDist)value.getValue()).getDisplayableName() +
				" (Right-click on background to change distribution.)");
	  }
	  
	  // If the panel is right-clicked, a popup with the allowable
	  // Distribution types should surface.  The popup also allows the
	  // selection of "no distribution" and the re-creation of the
	  // default distribution.
	  //
	  // NOTE: This is really ugly... the FD should really take
	  // care of some of this... somehow...
	  // 
	  // For popup menu stuff, see:
	  // http://java.sun.com/docs/books/tutorial/ui/swing/menu.html#popup
	  final JPopupMenu jpm = new JPopupMenu("Change Distribution");
	  final MouseAdapter ma = new MouseAdapter() {
	      public void mousePressed(MouseEvent e) {
		  maybeShowPopup(e);
	      }
	
	      public void mouseReleased(MouseEvent e) {
		  maybeShowPopup(e);
	      }
	  
	      private void maybeShowPopup(MouseEvent e) {
		  if (e.isPopupTrigger()) {
		      jpm.show(e.getComponent(),
			       e.getX(), e.getY());
		  }
	      }
	  };
	  // Add the default menu items: no dist [null value]
	  JMenuItem jmi = new JMenuItem("No Distribution");
	  jmi.addActionListener( new ActionListener() {
	      public void actionPerformed(ActionEvent ae) {
		  value.setValue( null );
		  jp.removeAll();
		  eipc = new JLabel("No Distribution");
		  jp.add(eipc);
		  jp.validate();
	      }
	  });
	  jpm.add(jmi);
	  // Add the default menu items: default dist [from the FD].
	  jmi = new JMenuItem("Default Distribution");
	  jmi.addActionListener( new ActionListener() {
	      public void actionPerformed(ActionEvent ae) {
		  value.setValue( fd.getDefaultValue() );
		  jp.removeAll();
		  // Setup the default distibution editor.
		  setupCustomEditor();
		  jp.validate();
	      }
	  });
	  jpm.add(jmi);
	  jp.addMouseListener(ma);
	  setupCustomEditor();
	  if (! (fd instanceof DiscreteFeatureDescriptor)) {
	      // Extend the popup menu created above.
	      jmi = new JMenuItem("Gaussian Dist");
	      jmi.addActionListener( new ActionListener() {
		  public void actionPerformed(ActionEvent ae) {
		      value.setValue( new GaussianDist() );
		      jp.remove(eipc);
		      eipc = ((ProbDist)value.getValue()).getEditInplacePanel();
		      jp.add(eipc);
		      jp.validate();
		  }
	      });
	      jpm.add(jmi);
	      jmi = new JMenuItem("Uniform Dist");
	      jmi.addActionListener( new ActionListener() {
		  public void actionPerformed(ActionEvent ae) {
		      value.setValue( new UniformDist() );
		      jp.remove(eipc);
		      eipc = ((ProbDist)value.getValue()).getEditInplacePanel();
		      jp.add(eipc);
		      jp.validate();
		  }
	      });
	      jpm.add(jmi);
	  }
	  toRet = (JComponent)jp;
      }

      toRet.validate();
      return (Component) toRet;
      
  }


    /**
     * This method actually creates the different kinds of custom
     * editors for the default distribution.
     * <P>
     * It handles these in two ways: 
     * <DL>
     * <DT>DiscreteFeatureDescriptors
     * <DD>The values are edited in a ComboBox, sorted by probability.
     * Selections in the ComboBox cause that probability to go to one,
     * and the others to zero.  Editing of the probability
     * distribution is available. 
     * 
     * <DT>Other FeatureDescriptor types
     * <DD>The values are edited with Component returned by the
     * ProbDist.getEditInplacePanel() method.
     * </DL>
     **/
    protected void setupCustomEditor()
    {
	if (fd instanceof DiscreteFeatureDescriptor &&
	    (value.getValue() != null) ) {
	    // Create a small list with the discrete elements sorted by
	    // probability.
	    // dfdcbm and jcb are final so the inner class can use them.
	    final DFDComboBoxModel dfdcbm = new DFDComboBoxModel();
	    final JComboBox jcb = new JComboBox( dfdcbm );
	    
	    GridBagLayout gbl = (GridBagLayout)jp.getLayout();
	    GridBagConstraints c = new GridBagConstraints();

      JButton jb = new JButton("Edit");
      jb.addActionListener( new ActionListener() {
	public void actionPerformed(ActionEvent a) {
	  Thread t = new Thread() {
	    public void run() {
	      // If "Edit" is clicked, create the table.
	      // JTable stuff with names and values.
	      GridBagLayout gblTmp = new GridBagLayout();
	      GridBagConstraints gbc = new GridBagConstraints();
	      JPanel jbp = new JPanel();
	      final JFrame jf = new JFrame();
	      jf.getContentPane().setLayout(gblTmp);

	      final DFDTableModel atm = new DFDTableModel();
	      final JTable jt = new JTable(atm);
	      jt.setPreferredScrollableViewportSize(new Dimension(300, 200));
	      JScrollPane jsp = new JScrollPane(jt);
	      gbc.gridwidth = GridBagConstraints.REMAINDER;
	      gblTmp.setConstraints(jsp, gbc);
	      jf.getContentPane().add(jsp);

	      JButton b = new JButton("Normalize");
	      b.addActionListener( new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		  // If editing, Stop table editing
		  if (jt.getCellEditor() != null) {
		    jt.getCellEditor().stopCellEditing();
		  }
		  ((DiscreteDist)value.getValue()).normalize();
		  // Update the table
		  atm.update();
		}
	      });
	      jbp.add(b);

	      b = new JButton("Dismiss");
	      b.addActionListener( new ActionListener() {
		public void actionPerformed(ActionEvent ae) {
		  // Close the window.
		  jf.dispatchEvent(new WindowEvent(jf,
						   WindowEvent.WINDOW_CLOSING));
		}
	      });
	      jbp.add(b);

	      gbc.gridwidth = GridBagConstraints.REMAINDER;
	      gbc.anchor = GridBagConstraints.CENTER;
	      gbc.fill = GridBagConstraints.NONE;
	      gblTmp.setConstraints(b, gbc);
	      jf.getContentPane().add(jbp);

	      WindowWaiter ww = new WindowWaiter();
	      jf.addWindowListener(ww);
	      jf.pack();
	      jf.setVisible(true);
	      ww.WaitUntilDone();
	      // When done, update the list.
	      dfdcbm.update();
	      jcb.validate();
	    }
	  }; // end new thread
	  t.start();
	}
      });
      c.gridwidth = GridBagConstraints.RELATIVE;
      c.fill = GridBagConstraints.HORIZONTAL;
      c.anchor = GridBagConstraints.WEST;
      gbl.setConstraints(jcb, c);
      jp.add(jcb);

      c.gridwidth = GridBagConstraints.REMAINDER;
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.EAST;
      gbl.setConstraints(jb, c);
      jp.add(jb);
    } else {
	if (value.getValue() == null) {
	    eipc = new JLabel( "No Distribution" ); 
	} else {
	    eipc = ((ProbDist)value.getValue()).getEditInplacePanel();
	}
      jp.add(eipc);
    }
  }

    public boolean isPaintable()
    {
	// This isn't used by the new PropertyCollectionSheetPanel.
	return false;
    }
    
    public void paintValue(Graphics gfx, Rectangle r)
    {
	; // no-op.  paintValue(...) shouldn't be used.
    }

    public boolean supportsCustomEditor()
    {
	// If probabilistic
	return fd.isProbabilistic();
    }
    
    /* INNER CLASS for dealing with the DiscreteFeatureDescriptor and
       editing it's values in a table. */

    /**
     * An AbstractTableModel subclass which provides a TableModel for
     * editing a probabilistic Discrete Feature.
     * <P>
     * This class also provides a method called <TT>update()</TT> which
     * fires the appropriate changes to get the table to update.
     * Unfortunately, the <TT>fireXXX(...)</TT> methods are protected,
     * so this lets clients get access to them.
     **/
    class DFDTableModel
	extends AbstractTableModel {
	private NumberFormat nf;
	
	public DFDTableModel() {
	    nf = NumberFormat.getNumberInstance();
	    nf.setMaximumFractionDigits(5);
	}
	
	public int getColumnCount() { return 2; }
	public int getRowCount() {
	    return ((DiscreteFeatureDescriptor)fd).getValueCount().intValue();
	}
	public String getColumnName(int col) {
	    String toRet = null;
	    switch (col) {
	    case 0:
		toRet = new String("Discrete Element");
		break;
	    case 1:
		toRet = new String("Probability mass");
		break;
	    }
	    return toRet;
	}
	public Class getColumnClass(int col) {
	    Class toRet = null;
	    switch (col) {
	    case 0:
		toRet = java.lang.String.class;
		break;
	    case 1:
		toRet = java.lang.Double.class;
		break;
	    }
	    return toRet;
	}
	public Object getValueAt(int row, int col) {
	    Object toRet = null;
	    switch (col) {
	    case 0:
		toRet = ((DiscreteFeatureDescriptor)fd).getValueName(new Integer( row ));
		break;
	    case 1:
		String strVal = nf.format(((DiscreteDist)value.getValue()).getProb( row ));
	toRet = new Double( strVal );
	break;
      }
      return toRet;
    }
    public boolean isCellEditable(int row, int col) {
      if (col > 0) {
	return true;
      }
      return false;
    }
    public void setValueAt(Object aValue, int row, int col) {
      DiscreteDist dd = (DiscreteDist)value.getValue();
      try {
	Number n = nf.parse((String)aValue);
	dd.setProb(row, n.doubleValue());
      } catch (NumberFormatException nfe) {
	JOptionPane.showMessageDialog(null, "Field cannot have the value " +
				      (String)aValue + ".");
      } catch (ParseException pe) {
	JOptionPane.showMessageDialog(null, "Field cannot have the value " +
				      (String)aValue + ".");
      } // end try-catch
    }

    public void update() {
      fireTableDataChanged();
    }
	      
  } // end DFDTableModel inner class

  /**
   * This class provides a ComboBoxModel in which the
   * probabilities in a Discrete Feature are listed in sorted order.
   * <P>
   * It also provides a public <TT>update()</TT> method, which wraps
   * access to the protected <TT>fireXXX(...)</TT> methods for client
   * use. 
   **/
  protected class DFDComboBoxModel
      extends AbstractListModel
      implements ComboBoxModel
  {
    // The size of the list we're dealing with.
    int size;

    // Maps sorted indices to their values in the DFD.
    private int map[];

    private DiscreteFeatureDescriptor dfd;
    private NumberFormat nf;

      private Object sel;

    public DFDComboBoxModel() {
      super();
      dfd = (DiscreteFeatureDescriptor)fd;
      nf = NumberFormat.getNumberInstance();
      nf.setMaximumFractionDigits(5);
      update();
    }

      public Object getSelectedItem()
      { return sel; }

      public void setSelectedItem(Object o)
      {
	  int idx = ((String)o).lastIndexOf(" (");
	  if (idx != -1) {
	      String valueName = ((String)o).substring(0, idx);
	      int newIdx = dfd.getValueIndex( valueName ).intValue();
	      // Set all values to 0.0
	      for (int i=0; i<dfd.getValueCount().intValue(); i++) {
		  ((DiscreteDist)value.getValue()).setProb( i,
							    0.0 );
	      } // end for over i
	      // Set the new value to 1.0
	      if (newIdx != -1) {
		  ((DiscreteDist)value.getValue()).setProb( newIdx,
							    1.0 );
	      } // end if
	      update();
	  } // end if
      }
	  
    public int getSize() { return size; }

    public Object getElementAt(int i) { 
 	return new String(dfd.getValueName(new Integer(map[i])) + " (" +
 			nf.format(((DiscreteDist)value.getValue()).getProb(map[i])) +
 			")" );
    };

    public void update() {
      size = dfd.getValueCount().intValue();
      DiscreteDist dd = (DiscreteDist)value.getValue();
      Vector v = new Vector( size );

      int insertIdx;
      // Get and sort the values [by probability] in the FD.
      for (int i=0; i<size; i++) {
	insertIdx = -1;
	// Do an insertion sort into the vector.
	for (int j=0; j<v.size() && (insertIdx == -1); j++) {
	  if ( dd.getProb(((Integer)v.elementAt(j)).intValue()) <=
	       dd.getProb(i) ) {
	    insertIdx = j;
	  } // end if
	} // end for over j

	if (insertIdx == -1) {
	  v.addElement(new Integer(i));
	} else {
	  v.insertElementAt(new Integer(i), insertIdx);
	} // end if-else
      } // end for over i

      // Create the map between list indices and value names.

      // ASSERT that v.size() == size
      if (v.size() != size) {
	throw new InternalError("Vector size != DFD size.");
      }
      map = new int[v.size()];
      for (int i=0; i < v.size(); i++) {
	map[i] = ((Integer)v.elementAt(i)).intValue();
      }

      // Set the selected index to the highest-probability item [at list position 0].
      sel = getElementAt(0);

      fireContentsChanged( this, 0, size );
    }
  } // end DFDComboBoxModel inner class
}
