package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import NTLSystem.*;
import NTLSystem.support.*;


public class FstructViewer
    extends JPanel
    implements ActionListener
{

    // {{{ Attributes 

    /**
     * Constraints for putting a collection of FstructViewers into a
     * GridBagLayout-based, JScrollPane-contained Panel for display.
     **/
    public static GridBagConstraints panelConstraints;

    /* Static init for panelConstraints */
    static {
	panelConstraints = new GridBagConstraints();
	panelConstraints.gridwidth = 1;
	panelConstraints.fill = GridBagConstraints.VERTICAL;
	panelConstraints.anchor = GridBagConstraints.NORTH;
	panelConstraints.weightx = panelConstraints.weighty = 1.0;
    }

    /**
     * The minimum height for the FstructViewer.
     **/
    private static int MINHEIGHT = 400;

    /**
     * The cached minimum necessary width.
     **/
    private transient int minWidth;
    /**
     * The cached minimum necessary height.
     **/
    private transient int minHeight;
	
    /**
     * The Fstruct being edited.
     **/
    private Fstruct fstruct;
    /**
     * The FeatureDesc object currently being used for the
     * PropertyCollectionSheetPanel.
     **/
    private Fstruct.FeatureDesc feature_desc;
    /**
     * The ignored features
     **/
    private String[] ignored;
    /**
     * The irrelevant features
     **/
    private String[] irrelevant;

    /**
     * The GridBagLayout for the Viewer.
     **/
    private GridBagLayout gbl;
    /**
     * The properties frame [when editing properties].
     **/
    private JFrame pf;
    /**
     * The name TextField [when editing].
     **/
    private JTextField nameTF;
    /**
     * The 'edit properties' button
     **/
    private JButton editProperties;
    /**
     * The 'add/remove features' button
     **/
    private JButton addRemoveFeatures;
    /**
     * The 'show irrelevant' checkbox.
     **/
    private JCheckBox irrelCB;
    /**
     * The Features editing panel [when editing].
     **/
    private PropertyCollectionSheetPanel feature_panel;

    // }}} Attributes

    // {{{ Constructors 

    /**
     * Create an Fstruct viewer for the given target Fstruct.
     *
     * @param     target The Fstruct to manipulate.  If null, the
     *            editing fields will be created, but not
     *            initialized. 
     **/
    public FstructViewer(Fstruct target)
    {
	this( target, null, null );
    }

    /**
     * Create an FstructViewer for the given target Fstruct, taking
     * into account the given ignored and irrelevant Features.
     *
     * @param     target The target Fstruct to manipulate.  If the
     *            target is 'null', fields will be created but will
     *            not be initialized.
     * @param     ignore An array of Feature names to ignore.
     * @param     irrelevant An array of Feature names which are
     *            'irrelevant'. 
     **/
    public FstructViewer(Fstruct target,
			 String[] ignore,
			 String[] irrelevant)
    {
	super();
	
	fstruct = target;

	ignored = ignore;
	this.irrelevant = irrelevant;

	gbl = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();
	setLayout(gbl);

	JPanel p;
	JButton b;
	JLabel l;

	// Top row with the name Text field
	c.weightx = c.weighty = 0.0;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.CENTER;

	c.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Name");
	gbl.setConstraints(l, c);
	add(l);

	c.anchor = GridBagConstraints.CENTER;
	c.gridwidth = GridBagConstraints.REMAINDER;
	c.fill = GridBagConstraints.HORIZONTAL;
	// Actual field init. is in 'setTarget(Fstruct)'
	nameTF = new JTextField( "", 15);
	gbl.setConstraints(nameTF, c);
	add(nameTF);

	minWidth = l.getPreferredSize().width + nameTF.getPreferredSize().width;
	minHeight = Math.max(l.getPreferredSize().height,
			     nameTF.getPreferredSize().height);

	c.gridwidth = GridBagConstraints.RELATIVE;
	editProperties = new JButton("Edit Properties");
	editProperties.setActionCommand("PROP_EDIT");
	editProperties.addActionListener( this );
	gbl.setConstraints(editProperties, c);
	add( editProperties );
	if (target == null) {
	    editProperties.setEnabled( false );
	}

	c.gridwidth = GridBagConstraints.REMAINDER;
	addRemoveFeatures = new JButton("Add/Remove Features");
	addRemoveFeatures.setActionCommand("FEATURE_ADD_REMOVE");
	addRemoveFeatures.addActionListener( this );
	gbl.setConstraints(addRemoveFeatures, c);
	add( addRemoveFeatures );
	if (target == null) {
	    addRemoveFeatures.setEnabled( false );
	}

	minWidth = Math.max( minWidth,
			     editProperties.getPreferredSize().width +
			     addRemoveFeatures.getPreferredSize().width );
	minHeight += Math.max( editProperties.getPreferredSize().height,
			       addRemoveFeatures.getPreferredSize().height );

	// Add ability to toggle display of relevant features.
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.CENTER;
	irrelCB = new JCheckBox("Show Irrelevant Features", true);
	irrelCB.addItemListener( new ItemListener() {
	    public void itemStateChanged(ItemEvent ie) {
		if (feature_desc != null) {
		    feature_desc.setIrrelevantFeatureShowing( irrelCB.isSelected() );
		    feature_panel.setTarget(feature_desc);
		    validate();
		}
	    }
	} );
	gbl.setConstraints(irrelCB, c);
	add(irrelCB);
	if (target == null) {
	    irrelCB.setEnabled( false );
	}

	minWidth = Math.max( minWidth, 
			     irrelCB.getPreferredSize().width );
	minHeight += irrelCB.getPreferredSize().height;

	c.weightx = c.weighty = 1.0;
	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = GridBagConstraints.REMAINDER;
	feature_panel = new PropertyCollectionSheetPanel();
	JScrollPane jsp = new JScrollPane(feature_panel,
					  ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
					  ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	gbl.setConstraints(jsp, c);
	add( jsp );

	minWidth = Math.max( minWidth,
			     jsp.getMinimumSize().width );
	minHeight += jsp.getMinimumSize().height;
	
	if (fstruct != null) {
	    setTarget( fstruct );
	}
    }

    // }}} Constructors 

    // {{{ Methods

    /**
     * Gets the target for this FstructViewer.
     *
     * @return    The 'target' for this FstructViewer.
     **/
    public Fstruct getTarget()
    {
	return fstruct;
    }

    /**
     * Sets the target which this FstructViewer is viewing.  This
     * initializes the appropriate GUI components, etc.  The 'ignored'
     * and 'irrelevant' arrays are kept from initialization.  They can
     * be changed with setIgnoredFeatures(String[]) and
     * setIrrelevantFeatures(String[]).
     *
     * @param     target The Fstruct to view.  This Fstruct cannot be
     *            null. 
     **/
    public void setTarget(Fstruct target)
    {
	// [Re]set the appropriate fields
	fstruct = target;
	
	if ( target != null ) {
	    nameTF.setText( fstruct.getName() );
	} else {
	    nameTF.setText( " [no fstruct] " );
	}

	updateFeaturePanel();

	// Enable functionality.
	addRemoveFeatures.setEnabled( true );
	editProperties.setEnabled( true );
	irrelCB.setEnabled( true );
    }

    /**
     * Sets the ignored Features for this FstructViewer.
     *
     * @param     ignored An array of String Feature names which are
     *            not to be shown in this FstructViewer.
     **/
    public void setIgnoredFeatures(String[] ignored)
    {
	this.ignored = ignored;
	// Update the GUI
	updateFeaturePanel();
    }

    /**
     * Sets the irrelevant Features for this FstructViewer.
     *
     * @param     irrel An array of String Feature names which are
     *            'irrelevant', and won't be shown in this
     *            FstructViewer depending on the state of the
     *            'irrelCB' check box. 
     **/
    public void setIrrelevantFeatures(String[] irrel)
    {
	this.irrelevant = irrel;
	// Update the GUI
	updateFeaturePanel();
    }

    /**
     * Update the Feature panel.  This involves getting a new
     * FeatureDesc object from the Fstruct, resetting the
     * PropertyCollectionSheetPanel's target to be this new
     * FeatureDesc object, and validating the layout of the whole
     * shoot-n-shebang. 
     **/
    protected void updateFeaturePanel()
    {
	if ( fstruct != null ) {
	    feature_desc = fstruct.getFeatureDesc( ignored,
						   irrelevant,
						   irrelCB.isSelected() );
	    feature_panel.setTarget( feature_desc );
	} else {
	    feature_panel.setTarget( null );
	}
	validate();
    }

    // {{{ ActionListener impl.

    public void actionPerformed(ActionEvent ae) {
	String aCmd = ae.getActionCommand();
	if (aCmd.equals("PROP_EDIT")) {
	    // If we can, bring up the properties editor.
	    JFrame pf =
		new PropertyCollectionSheet(fstruct.getPropertiesDesc(),
					    200, 300);
	    pf.pack();

	} else if (aCmd.equals("FEATURE_ADD_REMOVE")) {
	    // Manipulator features from the FeatureDictionary.
	    Thread t = new Thread() {
		public void run() {
		    JFrame edef = new EditableDataManipulator("Feature addition/removal",
							      fstruct,
							      fstruct.getDictionary() );
		    // Wait for the editor to finish.
		    WindowWaiter ww = new WindowWaiter();
		    edef.addWindowListener(ww);
		    ww.WaitUntilDone();
		    // Update the feature panel when done.
		    updateFeaturePanel();
		} 
	    };
	    t.start();
	}
    }

    // }}} ActionListener impl.


    public Dimension getMinimumSize()
    {
	return new Dimension(minWidth, Math.min(MINHEIGHT, minHeight) );
    }

    public Dimension getPreferredSize()
    {
	return new Dimension(minWidth, MINHEIGHT);
    }

    public void saveValues()
    {
	// Save the name TF... everything else is saved in-place.
	fstruct.setName( nameTF.getText() );
    }

    // }}} Methods

}


