package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.*;
import java.util.Random;

/**
 * A Gaussian distribution.  Gaussian Distributions are immutable.
 * <P>
 * $Id: GaussianDist.java,v 1.10 1998/10/16 01:57:10 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author Josh Sled [jsled@icsi.berkeley.edu]
 * @version $Revision: 1.10 $
 **/
public class GaussianDist 
implements ProbDist
{
  /**
   * The serialization Stream Unique Identifier.  Recompute this [with
   * serialver] when an incompatible change is made.
   **/
  static final long serialVersionUID = 2877155961717911435L;

  /* ATTRIBUTES */

  /** The mean of the distribution **/
  private double mean;
  /** The variance of the distribution **/
  private double variance;
  /** A PRNG for generating random values.  This is lazy-created when
   * needed. 
   * <P>
   * NOTE: This should be a "weak reference" so it can be gc'd if
   * necessary... 
   **/ 
  private Random prng;

  /**
   * A text field used for editing the mean [when editing].
   **/
  private JTextField meanTF;
  /**
   * A text field used for editing the variance [when editing].
   **/
  private JTextField varTF;


  /* CONSTRUCTORS */

  /**
   * Constructs a "default" GaussianDist, with a mean of 0.0 and an
   * variance of 1.0.
   * <P>
   * No-arg constructor for Serialization/Externalization.
   **/
  public GaussianDist()
  {
    super();
    setMean(0.0);
    setVariance(1.0);
  }

  /**
   * Constructs a new GaussianDist with the specified mean and
   * variance. 
   *
   * @param     mean The initial mean for the distribution.
   * @param     variance The initial variance for the distribution.
   **/
  public GaussianDist(double mean, double variance)
  {
    super();
    setMean(mean);
    setVariance(variance);
  }

  /* Distribution-specific value access methods */

  /**
   * Sets the mean value to the given mean.
   *
   * @param     newMean The new mean value to set.
   **/
  protected void setMean(double newMean)
  {
    mean = newMean;
  }

  /**
   * @return     The mean value for this GaussianDist.
   **/
  public double getMean()
  {
    return mean;
  }

  /**
   * Sets the variance.
   *
   * @param     newVariance The new variance for this distribution.
   **/
  protected void setVariance(double newVariance)
  {
    variance = newVariance;
  }

  /**
   * @return    The variance of this GaussianDist.
   **/
  public double getVariance()
  {
    return variance;
  }

  /* Implementation of the ProbDist interface */

  /**
   * Returns the probability at a specific value of interest in the
   * GaussianDist.
   * 
   * @param     fValue The value of interest.
   * @return    The probability at that value.
   **/
  public double getProb(double fValue)
  {
    // returns value of Gaussian at fValue 
    // NOTES: 
    // From: pg.733 AIMA:
    // G_{\sigma}(x) = (1 / sqrt(2 \pi) \sigma) e^( -(x - \mu)^2 / 2 \sigma^2)
    //
    // From: http://csep1.phy.ornl.gov/mc/node19.html:
    // f(x) = (1 / (2 \pi \sigma)^(.5)) e^(- (x-\mu)^2 / 2 \sigma^2)
    // Where \mu = mean and \sigma^2 = variance
    // However, it should be:

    // f(x) = (1 / (2 \pi)^.5 \sigma) e^(- (x - \mu)^2 / 2 \sigma^2)

    double sigma_squared = getVariance();
    double mu = getMean();

    return
      (1 / (Math.sqrt(2 * Math.PI * sigma_squared))) *
      Math.exp( - ( Math.pow((fValue - mu), 2) / (2 * sigma_squared) ));
   
  }

  /**
   * Randomly chooses a value from the distribution.
   *
   * @return     A randomly-chosen value from the distribution.
   **/
  public Object chooseValue()
  {
     Double wrapper;
    
    // Lazy-create the PRNG
    if (prng == null) {
      // Seed the PRNG
      prng = new Random(System.currentTimeMillis());
    }
    // Get a random "normally"-distributed value
    double randVal = prng.nextGaussian();

    // Scale it to this Gaussian dist., and wrap it in a Double
    wrapper = new Double((randVal * getVariance()) + getMean());
    return (Object)wrapper;
  }

  /**
   * Returns the maximum likelihood value from the distribution.
   * For gaussians, this is simply the mean.
   * @return    The maximum likelihood value of the distribution.
   **/
  public Object getMLValue()
  {
     Double wrapper;

     wrapper = new Double(mean);
     return (Object)wrapper;
  }

  /**
   * Redefinition of Object.equals(Object)
   **/
  public boolean equals(Object o)
  {
    // Do instance and null check
    if (! getClass().isInstance(o)) { return false; }
    // Cast for ease of reference
    GaussianDist gd = (GaussianDist)o;
    return ( (gd.getMean() == this.getMean()) &&
	     (gd.getVariance() == this.getVariance()) );
  }

  /**
   * "Normalizes" a GaussianDist, which does nothing.
   **/
  public void normalize()
  {
    ; // noop
  }

  /**
   * Reports on the distribution's continuity.
   *
   * @return    True, since GaussianDists are continuous.
   **/
  public boolean isContinuous()
  {
    return true;
  }

  /**
   * Redefinition of Object.toString()
   **/
  public String toString()
  {
    StringBuffer sb = new StringBuffer(30);
    sb.append("GaussianDist{");
    sb.append("mean=");
    sb.append(getMean());
    sb.append(", variance=");
    sb.append(getVariance());
    sb.append("}");
    return sb.toString();
  }

  /**
   * Redefinition of Object.clone()
   **/
  public Object clone()
  {
    try {
      GaussianDist toRet = (GaussianDist)super.clone();
      toRet.setMean(this.getMean());
      toRet.setVariance(this.getVariance());
      return toRet;
    } catch (CloneNotSupportedException cnse) {
      // Shouldn't happen because Cloneable is implemented.
      throw new InternalError();
    }
  }

  /**
   * Reads the GaussianDist from an ObjectInput implementer.
   * <P>
   * The format is specified in
   * <A HREF="#writeExternal(ObjectOutput)">writeExternal(...)</A> below.
   *
   * @param     input The ObjectInput object to read from.
   **/
  public void readExternal(ObjectInput input)
  {
    // FIXME
  }
  
  /**
   * Writes a GaussianDist in human-readable, human-editable format to
   * the given ObjectOutput.
   * <P>
   * Format is as follows:<BR>
   * FIXME
   *
   * @param     output The ObjectOutput object to write to.
   **/
  public void writeExternal(ObjectOutput output)
  {
    // FIXME
  }

  public String getDisplayableName()
  {
    return new String("Gaussian Dist.");
  }

  /**
   * @see ProbDist
   **/
  public Component getEditInplacePanel()
  {
    JPanel toRet = (JPanel)getEditViewPanel(true);
    KeyAdapter kl = new KeyAdapter() {
      public void keyTyped(KeyEvent ke) {
	try {
	  GaussianDist.this.saveValues();
	} catch (NotBeingEditedException nbee) {
	  // Won't happen.
	} catch (CannotSaveValuesException csve) {
	  // What to do, what to do?
	  // FIXME
	}
      }
    };
    meanTF.addKeyListener(kl);
    varTF.addKeyListener(kl);
    return toRet;
  }

  /**
   * Gets an edit or viewing panel from the distribution for
   * editing/viewing ability. 
   **/
  public Component getEditViewPanel(boolean edit)
  {
    JLabel l;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();

    JPanel toRet = new JPanel(gbl);
    
    meanTF = new JTextField(Double.toString(mean), 10);
    varTF = new JTextField(Double.toString(variance), 10);

    if (! edit) {
      meanTF.setEditable(false);
      varTF.setEditable(false);
    }

    gbc.fill = GridBagConstraints.NONE;
    gbc.gridwidth = GridBagConstraints.RELATIVE;
    l = new JLabel("Mean:");
    gbl.setConstraints(l, gbc);
    toRet.add(l);

    gbc.fill = GridBagConstraints.HORIZONTAL;
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbl.setConstraints(meanTF, gbc);
    toRet.add(meanTF);

    gbc.fill = GridBagConstraints.NONE;
    gbc.gridwidth = GridBagConstraints.RELATIVE;
    l = new JLabel("Variance:");
    gbl.setConstraints(l, gbc);
    toRet.add(l);

    gbc.fill = GridBagConstraints.HORIZONTAL;
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbl.setConstraints(varTF, gbc);
    toRet.add(varTF);

    return toRet;

  }

  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException
  {
    if ( (meanTF == null) || (varTF == null) ) {
      throw new NotBeingEditedException();
    }

    NumberFormat nf = NumberFormat.getNumberInstance();

    try {
      mean = nf.parse(meanTF.getText()).doubleValue();
      variance = nf.parse(varTF.getText()).doubleValue();
    } catch (ParseException pe) {
      throw new CannotSaveValuesException(pe.getMessage());
    } // end try-catch

  }

  public void doneEditing()
  {
    meanTF = varTF = null;
  }

}
