package NTLSystem.verblearn;

import java.io.*;
import java.util.*;
import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import NTLSystem.support.*;

///////////////////////////////////////////////////////////////////////
// A set of Scenarios.
// Note: Loading and saving routines do their own file handling, 
// unlike other classes in the VerbLearn system.

/**
 * A ScenarioSet contains 0..n Scenarios.
 * <P>
 * $Id: ScenarioSet.java,v 1.1 1998/07/28 23:58:07 nchang Exp $<BR>
 * Last editor: $Author: nchang $
 *
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @see       EditableDataDictionary
 * @see       Scenario
 * @version   $Revision: 1.1 $
 **/
public class Scenario Set
extends AbstractListModel
implements Serializable, Cloneable
{

  // serialization stream unique identifier
  // recompute [with serialver] when this class changes.
  static final long serialVersionUID = -5126178739094602841L;

  /* ATTRIBUTES */

  /** The actual data stored. **/
  protected Hashtable scenarios;  // maps String to Scenario

  /**
   * The keys used for the Hashtable, stored in sorted order.
   **/
  protected Vector scenarioNames;  // Vector<String>

  // these imported from original VL
  protected boolean labelsModified;
  protected LabelList[] labelList;   // nc added to keep track of labels

  private String specsFilename = "scenario/specs"; 

  /* GUI stuff */

  /**
   * The list of elements in the dictionary [when editing].
   **/
  private transient JList l;

  /**
   * The editing frame [when editing].
   **/
  private transient JFrame f;

  /* CONSTRUCTORS */

  /**
   * Basic constructor for a ScenarioSet.
   **/
  public ScenarioSet()
  {
    scenarios = new Hashtable();
    scenarioNames = new Vector();

    Hashtable goals = new Hashtable();    // maps String to Fstruct
    labelList = new LabelList[VerbLearn.context().numSlots()];

    //	scenarioNames.addElement(s);
    //	  scenarios.put(s, new Scenario(s, e, (Fstruct)goals.get(input.sval)));
    //	  scenarios.put(s, new Scenario(s, new Fstruct(initial),
    //					new Fstruct(input.sval)));
  }

  /**
   * Construct a ScenarioSet from a text file.
   **/
  {
    scenarioNames = new Vector();
    scenarios = new Hashtable();
    Hashtable goals = new Hashtable();    // maps String to Fstruct
    labelList = new LabelList[VerbLearn.context().numSlots()];

    NestedStreamTokenizer input;
    int token;
    try {
      // Create goal f-structs:
      input = DataSource.getNestedStreamTokenizer("scenario/goals");
      token = input.nextToken();
      while (token != NestedStreamTokenizer.TT_EOF) {
	if (token != NestedStreamTokenizer.TT_WORD) {
	  throw(new IOException("expected goal name in goals"));
	}
	String g = input.sval;
	token = input.nextToken();
	if (token != NestedStreamTokenizer.TT_NESTED) {
	  throw(new IOException("expected bracketed fstruct in goals"));
	}
	goals.put(g, new Fstruct(input.sval));
	token = input.nextToken();
      }

      // Create scenarios from specs:
      //input = DataSource.getNestedStreamTokenizer("scenario/specs");
      input = DataSource.getNestedStreamTokenizer(specsfile);
      token = input.nextToken();
      while (token != NestedStreamTokenizer.TT_EOF) {
	if (token != NestedStreamTokenizer.TT_WORD) {
	  throw(new IOException("expected scenario name in specs"));
	}
	String s = input.sval;
	scenarioNames.addElement(s);
	token = input.nextToken();
	if (token == NestedStreamTokenizer.TT_WORD) {
	  String e = input.sval;
	  token = input.nextToken();
	  if (token != NestedStreamTokenizer.TT_WORD) {
	    throw(new IOException("expected goal name in specs"));
	  }
	  if (!goals.containsKey(input.sval)) {
	    throw(new IOException("invalid goal name in specs"));
	  }
	  scenarios.put(s, new Scenario(s, e, (Fstruct)goals.get(input.sval)));
	}
	else if (token == NestedStreamTokenizer.TT_NESTED) {
	  String initial = input.sval;
	  token = input.nextToken();
	  if (token != NestedStreamTokenizer.TT_NESTED) {
	    throw(new IOException("expected final fstruct in fake scenario"));
	  }
	  scenarios.put(s, new Scenario(s, new Fstruct(initial),
					new Fstruct(input.sval)));
	}
	else {
	  throw(new 
		IOException("expected env name or nested fstruct in specs"));
	}
	token = input.nextToken();
      }

      // Add initial linking fstructs where available:
      String fname = "scenario/initial";
      if (DataSource.exists(fname)) {
	input = DataSource.getNestedStreamTokenizer(fname);
	token = input.nextToken();
	while (token != NestedStreamTokenizer.TT_EOF) {
	  if (token != NestedStreamTokenizer.TT_WORD) {
	    throw(new IOException("expected scenario name in initial links"));
	  }
	  String s = input.sval;
	  token = input.nextToken();
	  if (token == NestedStreamTokenizer.TT_NESTED) {
	    if (containsKey(s)) {
	      get(s).setInitialLink(new Fstruct(input.sval));
	    }
	  }
	  else {
	    throw(new IOException("expected nested fstruct in initial links"));
	  }
	  token = input.nextToken();
	}
      }
      
      // Add final linking fstructs where available:
      fname = "scenario/final";
      if (DataSource.exists(fname)) {
	input = DataSource.getNestedStreamTokenizer(fname);
	token = input.nextToken();
	while (token != NestedStreamTokenizer.TT_EOF) {
	  if (token != NestedStreamTokenizer.TT_WORD) {
	    throw(new IOException("expected scenario name in final links"));
	  }
	  String s = input.sval;
	  token = input.nextToken();
	  if (token == NestedStreamTokenizer.TT_NESTED) {
	    if (containsKey(s)) {
	      get(s).setFinalLink(new Fstruct(input.sval));
	    }
	  }
	  else {
	    throw(new IOException("expected nested fstruct in final links"));
	  }
	  token = input.nextToken();
	}
      }
    }
    catch (IOException e) {
     // System.err.println("ScenarioSet failed to load: " + e.getMessage());
      scenarioNames = new Vector();
      scenarios = new Hashtable();
    }
  }


  /* METHODS */

  // adding/removing Scenarios

  /**
   * Adds a scenario with specified name to ScenarioSet. 
   *   If this replaces an entry in the ScenarioSet, so be it.
   *
   * @param     sc The Scenario to put in the
   *            ScenarioSet.
   * @param     id The name of the Scenario to put in the ScenarioSet.
   **/
  public void addScenario(String id, Scenario sc) {
    scenarios.put(id, sc);
    if(!scenarioNames.contains(id)) scenarioNames.addElement(id);
  }
  
  /**
   * Adds a scenario to the ScenarioSet.   If this replaces an
   * entry in the ScenarioSet, so be it.
   *
   * @param     sc The Scenario to put in the
   *            ScenarioSet.
   **/
  public void addScenario(Scenario sc) {
    addScenario(sc.getKey(), sc);
  }

  /**
   * Removes a scenario from the ScenarioSet.  
   *
   * @param     sc The Scenario to remove from the
   *            ScenarioSet.
   **/
  public void removeScenario(String id){
      scenarioNames.removeElement(id);
      scenarios.remove(id);
  }
  
  /**
   * Removes a scenario from the ScenarioSet.  
   *
   * @param     sc The Scenario to remove from the
   *            ScenarioSet.
   **/
  public void removeScenario(Scenario sc){
      removeScenario(sc.id());
  }


  // retrieving Scenarios/names

  /**
   * Test to see if a particular named Scenario is in the
   * ScenarioSet.
   *
   * @param     searchName The name of the Scenario to test
   *            for.
   * @return    True if the Scenario is in the
   *            ScenarioSet, false if not.
   **/
  public boolean containsScenario(String searchName)
  {
    return scenarios.containsKey(searchName);
  }

  /**
   * Test to see if the particular Scenario object is in the
   * ScenarioSet.
   *
   * @param     otherSc The Scenario to check for.
   * @return    true If the Scenario is in the
   *            ScenarioSet, false if not.
   **/
  public boolean containsScenario(Scenario otherSc)
  {
    return scenarios.contains(otherSc);
  }
    
  /**
   * Gets the Scenario with the given name from the
   * ScenarioSet. Returns null if the object does not exist in
   * the ScenarioSet.
   *
   * @param     searchName The name of the Scenario to search
   *            for.
   * @return    The Scenario with the given name, or null if
   *            it does not exist.
   **/
  public Scenario getScenario(String searchName)
  {
    return (Scenario)scenarios.get(searchName);
  }

  /**
   * Returns an Enumeration of all the Scenarios in the
   * ScenarioSet. 
   *
   * @return    An Enumeration of all the Scenarios.
   **/
  public Enumeration getAllScenarios()
  {
    return scenarios.elements();
  }

  /**
   * Returns an enumeration of names of Scenarios in ScenarioSet
   *
   * @return    An enumeration of names of Scenarios in the ScenarioSet. 
   **/
  public Enumeration names()
    {
      return scenarioNames.elements();
    }

  /**
   * Returns a Vector of all the Scenarios names in the
   * ScenarioSet. 
   *
   * @return    A Vector of the Scenario names in the ScenarioSet.
   **/
  public Vector scenarioNames()
  {
    return scenarioNames;
  }

  /**
   * Returns the count of the number of Scenarios in the
   * ScenarioSet
   *
   * @return    The number of Scenarios in the
   *            ScenarioSet. 
   **/
  public int countScenarios()
  {
    return scenarios.size();
  }

  // label handling: is this needed? Maybe just word list?
  // should this become an editable dictionary? (actual lexicon is,
  // but is it necessary for this stage? 
 

  /**
   * Returns a LabelList (an array of label Vectors) of ScenarioSet.
   *
   * @return    LabelList of labels used in Scenarios.
   **/
  public LabelList getLabels(int slotNum)
    {
      return labelList[slotNum];
    }

  /**
   * Returns the count of the number of Scenarios in the
   * ScenarioSet
   *
   * @return    The number of Scenarios in the
   *            ScenarioSet. 
   **/
  public boolean labelsModified() { return labelsModified; }


  /**
   * Returns the count of the number of Scenarios in the
   * ScenarioSet; assumes sc is in the set.
   *
   * @return    The number of Scenarios in the
   *            ScenarioSet. 
   **/
  public void setLabel(Scenario sc, VerbComplex vc) {
    sc.setLabel(vc);
    setLabelsModified();
  }

  // Assumes sc is in the set.
  public void setInitialAndFinalLinks(Scenario sc, Fstruct i, Fstruct f) {
    sc.setInitialLink(i);
    sc.setFinalLink(f);
  }

  private void setLabelsModified() { 
    labelsModified = true; 
    setChanged(); notifyObservers(); VerbLearn.frame().validate();
  }

  public void loadLabels() {
    clearLabels();
    try {
      String fname = "lang/" + VerbLearn.context().langslot() +"/labels";
      if (DataSource.exists(fname)) {
	NestedStreamTokenizer input = 
	  DataSource.getNestedStreamTokenizer(fname);
	int token = input.nextToken();
	while (token != NestedStreamTokenizer.TT_EOF) {
	  if (token != NestedStreamTokenizer.TT_WORD) {
	    throw(new IOException("expected scenario name in labels"));
	  }
	  String s = input.sval;
	  token = input.nextToken();
	  if (token == NestedStreamTokenizer.TT_WORD) {
	    // nc added following to allow "(none)" to indicate no label
	    if (input.sval.equals("(none)")) {
	      get(s).setLabel(null);
	    }
	    else if (containsKey(s)) {
	      get(s).setLabel(new VerbComplex(input.sval));
	    }
	  }
	  else {
	    throw(new 
		  IOException("expected verb-complex in labels"));
	  }
	  token = input.nextToken();
	}
      }
    }
    catch (IOException e) {
      System.err.println("ScenarioSet loadLabels exception: " + 
			 e.getMessage());
      clearLabels();
    }
    setChanged(); notifyObservers(); VerbLearn.frame().validate();
  }

  // Return true if successful.
  public boolean saveLabels() {
    if (labelsModified) {
      try {
	String pathname = VerbLearn.DATA_BASE_DIR + "lang/" +
	  VerbLearn.context().langslot() + "/labels";
	DataOutputStream os = new 
	  DataOutputStream(new FileOutputStream(pathname));
	os.writeBytes("// Labels for " + VerbLearn.context().langslot() +
		      ":\n\n");
	Enumeration scnEnum = names();
	while (scnEnum.hasMoreElements()) {
	  Scenario sc = get((String)scnEnum.nextElement());
	  if (sc.isLabelled()) {
	    os.writeBytes(sc.id() + "\t" + sc.label() + "\n");
	  }
	}
	labelsModified = false;
	return true;
      }
      catch (IOException e) {
	return false;
      }
    }
    else {
      return true;
    }
  }

  private void clearLabels() {
    Enumeration scEnum = scenarios.elements();
    while (scEnum.hasMoreElements()) {
      ((Scenario)scEnum.nextElement()).setLabel(null);
    }
  }

  public void newLabelList() { 
    int num = VerbLearn.context().numSlots();
    labelList = new LabelList[num];
    for (int i =0; i < num; i++) {
      labelList[i] = new LabelList();
    }
  }

  public void clearLabelList() {
    labelList = null;
  }

  public boolean okLabelList() {
    return (labelList != null);
  }

  // if context changes, redo label list; no!!
  public void redoLabelList() {
    labelList = null;
    labelList = new LabelList[VerbLearn.context().numSlots()];
  }

  public String toString() {
    StringBuffer res = new StringBuffer().append("Scenario Set Summary:\n");
    Enumeration scnEnum = names();
    while (scnEnum.hasMoreElements()) {
      Scenario sc = get((String)scnEnum.nextElement());
      res.append("\nScenario ").append(sc.id()).append(":\n");
      if (sc.isGrounded()) {
	res.append("env = ").append(sc.env()).append("\n");
	res.append("goal = ").append(sc.goal());
      }
      if (sc.hasLinks()) {
	res.append("initial = ").append(sc.initialLink());
	res.append("final = ").append(sc.finalLink());
      }
      if (sc.isLabelled()) {
	res.append("label = ").append(sc.label()).append("\n");
      }
    }
    return res.toString();
  }

  // Incorporate a Vector of Scenarios (from the Generate panel).
  // If any of them have id's equal to an existing scenario,
  // OVERWRITE the current scenario, but keep its current label.
  // BEWARE:  This may be undesired behavior, particularly if the
  // user has done a "compact names" operation in the Generate panel.
  public void incorporate(Vector sv) {
    Enumeration sEnum = sv.elements();
    while (sEnum.hasMoreElements()) {
      Scenario s = (Scenario)sEnum.nextElement();
      if (containsKey(s.id())) {
	// Overwrite, after extracting current label
	Scenario cur = get(s.id());
	s.setLabel(cur.label());
	scenarios.put(s.id(), s);
      }
      else {
	// Add scenario to set
	scenarioNames.addElement(s.id());
	scenarios.put(s.id(), s);
      }
    }
    setChanged(); notifyObservers(); 
  }

  // merges another scenario set into this one
  public void merge(ScenarioSet ss){
    Enumeration sEnum = ss.names();
    while (sEnum.hasMoreElements()) {
      String name = (String)sEnum.nextElement();
      Scenario scen = ss.get(name);
      
      // if-else block could be relaced by next two lines:
      // if (containsKey(name)) scen.setLabel(get(name).label());
      // addScenario(scen);

      if (containsKey(name)) {
	// Overwrite, after extracting current label
	   Scenario cur = get(name);
	   scen.setLabel(cur.label());
	   scenarios.put(name, scen);
      }
      else {
	  scenarioNames.addElement(name);
	  scenarios.put(name, scen);
      }
    }
  }

  // Should be called after removing or restoring features in 
  // VerbLearn.context().params().  Within each fstruct in the
  // ScenarioSet, rearranges features between the "on" and "off"
  // feature Hashtables.
  public void rearrangeFstructs() {
    Enumeration scEnum = scenarios.elements();
    while (scEnum.hasMoreElements()) {
      Scenario sc = (Scenario)scEnum.nextElement();
      sc.initialLink().rearrangeFeatures();
      sc.finalLink().rearrangeFeatures();
    }
  }

  /**
   * Testing code. -- fix! -nc
   **/
  public static void main(String args[])
  {
    // Get the properties.
    String propFN = new String("properties/default.properties");
    Properties ourProps = new Properties(System.getProperties());
    try {
      FileInputStream propsIn = new FileInputStream(propFN);
      ourProps.load(propsIn);
      System.setProperties(ourProps);
    } catch (SecurityException se) {
      // Shouldn't happen as an app.
      throw new InternalError(se.getMessage());
    } catch (FileNotFoundException fnfe) {
      System.out.println("Could not find NTLSystem properties file " + propFN);
    } catch (IOException ioe) {
      System.out.println("Could not load default properties file.");
    } // end try-catch

    String scenSetName = ourProps.getProperty("ntlsys.dataDir") +
      "/" + ourProps.getProperty("ntlsys.dictDir") +
      "/" + ourProps.getProperty("ntlsys.scenSet");
    ScenarioSet ss = null;

    try {
      FileInputStream fis =
	new FileInputStream(scenSetName);
      ObjectInputStream ois = new ObjectInputStream(fis);
      ss = (ScenarioSet)ois.readObject();
    } catch (IOException ioe) {
      ioe.printStackTrace();
      System.out.println(ioe.getMessage());
    } catch (ClassNotFoundException cnfe) {
      // Won't happen
    }
    
    if (ss == null) {
      System.out.println("Scenario Set _not_ restored.");
      ss = new ScenarioSet();
    }
    ss.Edit(true);

    try {
      FileOutputStream fos = new FileOutputStream(dictFN);
      ObjectOutputStream oos = new ObjectOutputStream(fos);
      oos.writeObject(ss);
      oos.flush();
      fos.close();
    } catch (IOException ioe) {
      ioe.printStackTrace();
      System.out.println(ioe.getMessage());
    }

    System.exit(0); 
 }
}
