package NTLSystem.verblearn;

import java.io.*;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;
import NTLSystem.support.NestedStreamTokenizer;

/**
 * A Dataset is a named tuple of collections of Scenarios divided into
 * three groups:
 * <DL>
 * <DT>train
 * <DD>The Scenarios to be used for training.
 * <DT>testing
 * <DD>The Scenarios to be used for testing [recognition/obeying].
 * </DL>
 * <P>
 * It is not [necessarily] required that these be disjoint sets.
 * <P>
 * $Id: Dataset.java,v 1.4 1999/08/03 21:02:58 nchang Exp $<BR>
 * Last editor: $Author: nchang $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Word
 * @see       Lexicon
 * @see       DatasetEditor
 * @version   $Revision: 1.4 $
 **/
public class Dataset
    implements Serializable
{
  static final long serialVersionUID = -2983249527502760505L;
    /**
     * A Vector<Scenario> against which to train.
     **/
    private Vector train;
    /**
     * A Vector<Scenario> against which to test.
     **/
    private Vector test;
    /**
     * The name of this Dataset.
     **/
    private String name;

    /**
     * No-arg constructor for serialization.
     **/
    public Dataset() { }

    /**
     * A basic constructor which simply sets the name and initializes
     * the other fields to empty Vectors.
     *
     * @param     name The name of this Dataset.
     **/
    public Dataset(String name)
    {
	this.name = name;
	train = new Vector();
	test = new Vector();
    }

    /**
     * Basic, simple constructor... give it all the data it needs, and
     * it goes to work.
     *
     * @param     name The name of this Dataset.
     * @param     train The Vector of Scenario objects to be used for
     *            training.
     * @param     test The Vector of Scenario objects to be used for
     *            testing.
     **/
    public Dataset(String name,
		   Vector train,
		   Vector test)
    {
	this.name = name;
	this.train = train;
	this.test = test;
    }

    /**
     * Creates a Dataset from the given Enumeration of Scenarios,
     * based on the given percentage values.  Items in a set are
     * unique, and a flag to ensure disjoint sets is provided.
     *
     * @param     name The name of this Dataset.
     * @param     scenarios An Enumeration of the Scenarios to place
     *            in the Dataset.
     * @param     trainWeight The percentage of the elements to put in
     *            the training set.
     * @param     testWeight The percentage of the elements to put in
     *            the test set.
     * @param     disjoint True if the sets should be disjoint.
     **/
    public Dataset(String name, Enumeration scenarios,
		   double trainWeight, double testWeight,
		   boolean disjoint)
    {
	this(name);
	setsFromScenarios(scenarios,
			  trainWeight, testWeight,
			  false, // no, don't replace... the Vectors
			  // are empty at this point, anyways...
			  disjoint);
    }


    /** nc changes to this
     * Creates a Dataset from a dataset file of original VerbLearn format.
     *
     * @param     name The name of this Dataset.
     * @param     input Input file in original VerbLearn format.
     * @param     sset original Sset? probably null.
     **/
    public Dataset(String name,
		   NestedStreamTokenizer input,
		   ScenarioSet sset)
    {
	this.name = name;
	train = new Vector();
	test = new Vector();

	Scenario s;

	try {
	    int token = input.nextToken();
	    while (token != NestedStreamTokenizer.TT_EOF) {
		if (token != NestedStreamTokenizer.TT_WORD ||
		    (!input.sval.equals("train") && 
		     !input.sval.equals("recognize") &&
		     !input.sval.equals("obey"))) {
		    throw new IOException("expected train/recognize/obey");
		}
		Vector v = null;
		// Setup the approrpriate vector.
		if (input.sval.equals("train")) { v = train; }
		else if (input.sval.equals("recognize")) { v = test; }
		else if (input.sval.equals("obey")) { v = new Vector(); }

		token = input.nextToken();
		if (token != NestedStreamTokenizer.TT_NESTED) {
		    throw new IOException("expected nested list of scenarios");
		}
		NestedStreamTokenizer list =
		    new NestedStreamTokenizer(new StringReader(input.sval));
		token = list.nextToken();
		while (token != NestedStreamTokenizer.TT_EOF) {
		    String scName = list.sval;
		    String scNum = scName.substring(2);
		    switch ( scNum.length() ) {
		    case 1:
		      scNum = "00" + scNum;
		      break;
		    case 2:
		      scNum = "0" + scNum;
		      break;
		    }
		    scName = "sc" + scNum;
		    s = sset.getScenario( scName );
		    if ( s == null ) {
		      throw new IOException(" Invalid scenario name: " +
					    list.sval );
		    }
		    v.addElement( s );
		    token = list.nextToken();
		} // end while over scenario names.
		token = input.nextToken();
	    } // end while over dataset types.
	} catch (IOException e) {
	    System.err.println("Dataset load-from-file exception: " + 
			       e.getMessage());
	} // end try-catch
    } 

  // nc temp -- toString in OLD format, where recognize/obey forced to be same set
    public String toString() {
      StringBuffer res = new StringBuffer();
      res.append("// Dataset: ").append(name).append("\n\ntrain {");
      for (int i = 0; i < train.size(); i++) {
	res.append("\n\t").append(RemoveZeros(((Scenario) train.elementAt(i)).getName()));
      }
      res.append("\n}\nrecognize {");
      for (int i = 0; i < test.size(); i++) {
	res.append("\n\t").append(RemoveZeros(((Scenario) test.elementAt(i)).getName()));
      }
      res.append("\n}\nobey {");
      for (int i = 0; i < test.size(); i++) {
	res.append("\n\t").append(RemoveZeros(((Scenario) test.elementAt(i)).getName()));
      }
      res.append("\n}\n");
      return res.toString();
    }

    public String RemoveZeros(String old) {
      String temp = old.substring(2);
      int i = 0;
      while (temp.charAt(i) == '0')
	i++;
      String newSt = "sc" + temp.substring(i);
      return newSt;
    }

    /**
     * Get the name of this Dataset.
     *
     * @return    The name of this Dataset.
     **/
    public String getName()
    { return name; }

    /**
     * Sets the name of this Dataset.
     *
     * @param     newName The new name of this Dataset.
     **/
    public void setName(String newName)
    { name = newName; }

    /**
     * Get the collection of Scenario objects to be used for
     * training. 
     *
     * @return    An Enumeration<Scenario> designated for training.
     **/
    public Enumeration trainEnum() { return train.elements(); }

    /**
     * Sets the collection of Scenarios to be used for training.
     *
     * @param     train A Vector<Scenario> to be used for training.
     **/
    public void setTrainingScenarios(Vector train)
    { this.train = train; }

    /**
     * Get the collection of Scenario objects to be used for
     * testing.
     *
     * @return    An Enumeration<Scenario> designated for
     *            testing. 
     **/
    public Enumeration testEnum()
    { return test.elements(); }

    /**
     * Sets the collection of Scenarios to be used for testing.
     *
     * @param     testing A Vector<Scenario> to be used for
     *            testing. 
     **/
    public void setTestingScenarios(Vector testing)
    { this.test = testing; }

    /**
     * Replaces or adds to the sets in this Dataset from the given
     * Enumeration of Scenarios.  The given trainWeight and
     * testWeights are used as precentages to determine how many
     * scenarios [randomly chosen from the scenarios Enumeration] are
     * put in the sets.
     *
     * @param     scenarios An Enumeration of the Scenarios to
     *            process. 
     * @param     trainWeight The percentage of scenarios which the
     *            train set should contain.
     * @param     testWeight The percentage of scenarios which the
     *            test set should contain.
     * @param     replace If true, the data in the sets will be
     *            replace.  If false, they will be added to
     *            [uniqueness will be preservered].
     * @param     disjoint If true, scenarios added to the training
     *            set will not be added to the testing set.  If false,
     *            there is no such constraint, and each Scenario has a
     *            chance to be added to both the training and testing
     *            sets.
     **/
  public void setsFromScenarios(Enumeration scenarios,
				double trainWeight,
				double testWeight,
				boolean replace,
				boolean disjoint)
    {
      Random prng = new Random(System.currentTimeMillis());
      Scenario s;
      Vector scens = new Vector();
      
      // Move Scenarios from Enumeration to Vector.
      while ( scenarios.hasMoreElements() ) {
	scens.addElement( scenarios.nextElement() );
      }
      
      // Check to make sure they're not asking for more than we can
      // provide.
      if (disjoint && (trainWeight + testWeight > 1.0)) {
	return;
      }
      
      // Figure out how many of each we need.
      int trainCount, testCount;
      trainCount = (int)Math.floor( scens.size() * trainWeight );
      testCount = (int)Math.ceil( scens.size() * testWeight );
      
      if (replace) {
	train = new Vector( trainCount );
	test = new Vector( testCount );
      } else {
	trainCount = train.size();
	testCount = test.size();
      }
      
      // Add enough scenarios to training set.
	while ( trainCount > 0 ) {
	  // randomly choose element from scens
	  s = (Scenario)scens.
	    elementAt( (int)Math.floor( prng.nextDouble() * scens.size() ));
	  // make sure it's not already in training
	  if ( !train.contains( s ) ) {
	    if ( disjoint ) {
		    // if disjoint, make sure it's not in the test
	      // set.
	      if ( !test.contains( s ) ) {
		train.addElement( s );
		trainCount--;
	      }
	    } else {
		    train.addElement( s );
		    trainCount--;
		  }
	  }
	} // end while [adding to training set]

      // Add enough scenarios to testing set.
      while ( testCount > 0 ) {
	s = (Scenario)scens.
	  elementAt( (int)Math.floor(prng.nextDouble() * scens.size() ));
	// make sure it's not already in testing
	if ( !test.contains( s ) ) {
	  if ( disjoint ) {
	    if ( !train.contains( s ) ) {
	      test.addElement( s );
	      testCount--;
	    }
	  } else {
	    test.addElement( s );
	    testCount--;
		 }
	}
      } // end while [adding to testing set]
    }

  /**
   * Replaces or adds to the sets in this Dataset from the given
   * Enumeration of Scenarios.  The given trainWeight and
   * testWeights are used as precentages to determine how many
   * scenarios [randomly chosen from the scenarios Enumeration] are
   * put in the sets.
   *
   * @param     scenarios An Enumeration of the Scenarios to
   *            process. 
   * @param     testWeight The percentage of scenarios which the
   *            test set should contain.
   **/
  public void batchOutput(Vector training, Vector testing,
			  double batchPercent)
    {
      //	Random prng = new Random(System.currentTimeMillis());
      Scenario s;
      
      // Check to make sure they're not asking for more than we can
      // provide.
      if (batchPercent > 1.0) {
	return;
      }
      
      // Figure out how many in each batch
      int batchSize = (int)Math.floor( training.size() * batchPercent );
      System.err.println("batchSize is " + batchSize);
      
      // Write out to batches.
      
      int scenInd = 0;
      int batchNum = 0;
      Vector batch;
      int b;
      
      // Add enough scenarios to training set.
      while ( scenInd < training.size() ) {
	
	batch = new Vector( batchSize );
	for (b = 0; 
	     (b < batchSize) && (scenInd < training.size()); 
	     b++) {
	  s = (Scenario)training.elementAt(batchNum * batchSize + b);
	  batch.addElement(s);
	  scenInd++;
	}
	// save this as new file.
	Dataset newDataset = new Dataset(this.name + "-" + batchNum,
					 batch, test);
	String dir = "./data/datasets/";
	File toSave = new File(dir + this.name + "-" + batchNum);
	newDataset.save(toSave);
	batchNum++;
      } // end while [adding to training set]

    }

  // assumes file is valid input.
  public void save( File file) {
    try {
      PrintWriter pw = new PrintWriter(new FileWriter(file));
      pw.write(this.toString());
      pw.close();
    }
    catch (IOException e) {
	    System.err.println("Dataset failed to save exception: " + 
			       e.getMessage());
    }
  }
  
}


