package NTLSystem.verblearn

import java.io.*;
import java.util.*;
import NTLSystem.support.*;


///////////////////////////////////////////////////////////////////////
// Contains all words occurring in a single grammatical position, or
// "slot".  Also contains a local copy of the number of virtual samples
// to use for each feature, to allow slot-specific adaption.

class Slot
    implements Serializable
{

  private int id;

  public int id() { return id; }

  private Hashtable words;  // maps String version to its Word object

  public Hashtable virtuals; // maps String feature names to Double

  private double numEx;  // number of training examples in this slot

  Slot(int id) {
    this.id = id;
    words = new Hashtable();
    // FIXME: This should really be virtuals.clone();
    // FIXME: Where are virtuals to be stored?
    virtuals = VerbLearn.context().params().copyOfVirtuals();
  }

  // Return Word corresponding to 'w', or null if none exists.
  public Word get(String w) { return (Word)words.get(w); }

  public String[] wordList() {
    String[] wl = new String[words.size()];
    int i = 0;
    Enumeration kEnum = words.keys();
    while (kEnum.hasMoreElements()) {
      wl[i] = (String)kEnum.nextElement();
      i = i + 1;
    }
    return wl;
  }

  // Update slot to include a new training instance, adding a new Word
  // if necessary.
  public void incorporate(Fstruct f, String w) {

    numEx++;

    // Add a new Word if necessary:
    if (!words.containsKey(w)) {
      words.put(w, new Word(w, virtuals, this));
    }

    // Modify the Word to include the new usage:
    boolean ranMerge = ((Word)words.get(w)).incorporate(f);

    // Update the slot-specific virtual samples:
    if (VerbLearn.context().params().adaptVirtuals() && ranMerge) {
      adaptVirtuals();
    }
  }

  public void synch() {
    Enumeration wEnum = words.elements();
    while (wEnum.hasMoreElements()) {
      ((Word)wEnum.nextElement()).synch();
    }
    if (VerbLearn.context().params().adaptVirtuals()) {
      adaptVirtuals();
    }
  }

  // For each feature, do a weighted average of (1) the current
  // settings, and (2) the numEx-averaged value of v = (1-p)/(pn-1)
  // across all senses (where n is the # of values, and p is the
  // probability of the mode value).
  private void adaptVirtuals() {
    // Iteratively adapt each feature:
    Enumeration fEnum = virtuals.keys();
    while (fEnum.hasMoreElements()) {
      String fname = (String)fEnum.nextElement();
      double vAvg = 0;
      double count = 0;
      // Double loop to average mode's probability over all senses in slot:
      Enumeration wEnum = words.elements();
      while (wEnum.hasMoreElements()) {
	Enumeration sEnum = ((Word)wEnum.nextElement()).senseEnum();
	while (sEnum.hasMoreElements()) {
	  Sense s = (Sense)sEnum.nextElement();
	  double p = s.getDist(fname).likelihood(s.getDist(fname).mode());
	  // Avoid infinity when adaptVirtuals encounters a uniform dist:
	  vAvg = vAvg + s.numEx() * 
	    Math.min(VerbLearn.context().params().maxVirtual(),
	     (1 - p) / 
	     (p * VerbLearn.context().params().featureVals(fname).size() - 1));
	  count = count + s.numEx();
	}
      }
      // Weight the above against the current virtuals setting:
      vAvg = vAvg + VerbLearn.context().params().virtualInertia() *
	((Double)virtuals.get(fname)).doubleValue();
      count = count + VerbLearn.context().params().virtualInertia();
      vAvg = vAvg / count;
      // Finally set the new value:
      virtuals.put(fname, new Double(vAvg));
    }
  }

  public CandidateSense label(Fstruct f) { 
    CandidateSense leading = new CandidateSense();
    leading.posterior = -1;
    Enumeration wEnum = words.elements();
    while (wEnum.hasMoreElements()) {
      Word w = (Word)wEnum.nextElement();
      CandidateSense cur = w.bestSenseFor(f, numEx);
      if (cur.posterior > leading.posterior) {
	leading = cur;
	leading.label = w.id();
      }
    }
    return leading;
  }

  // Return posteriors of all word senses.
  public Vector allSensePosteriors(Fstruct f) {
    Vector res = new Vector();  // Vector<CandidateSense>
    Enumeration wEnum = words.elements();
    while (wEnum.hasMoreElements()) {
      Vector cands = ((Word)wEnum.nextElement()).allSensePosteriors(f, numEx);
      Enumeration cEnum = cands.elements();
      while (cEnum.hasMoreElements()) {
	res.addElement((CandidateSense)cEnum.nextElement());
      }
    }
    return res;
  }

  // Return the Sense for word "cmd" which best fits the world "state",
  // or null if the best fit's posterior isn't >= minObey or if
  // the word is unknown.
  public Sense obey(String cmd, Fstruct state) {
    if (cmd == null) return null;
    Word w = (Word)words.get(cmd);
    if (w == null) return null;
    CandidateSense cand = w.bestSenseFor(state, w.numEx());
    if (cand.posterior >= VerbLearn.context().params().minObey()) {
      return cand.sense;
    } else {
      return null;
    }
  }

  public String toString() {
    StringBuffer res = new StringBuffer();
    res.append("Slot ").append(id).append(" (").append(totalSenses())
      .append(" senses) virtuals: ");
    Enumeration fEnum = VerbLearn.context().params().featureNames();
    boolean firstTime = true;
    while (fEnum.hasMoreElements()) {
      String fname = (String)fEnum.nextElement();
      if (firstTime) {
	firstTime = false;
      }
      else {
	res.append(" ");
      }
      //      res.append(fname).append("=")
      //	.append(Fmt.dbl(((Double)virtuals.get(fname)).doubleValue()));
      // NC commented out above -- problems when displaying with new 
      // param set that may have different features on/off, since
      // virtuals doesn't nec'y have fname in it. 
      // for now replace with "none" -- maybe if changed
      // shouldn't even be displaying lexicon??
      res.append(fname).append("=");
      Double virtVal = (Double)(virtuals.get(fname));
      if (virtVal == null) {
	res.append("none");
      }
      else {
	res.append(Fmt.dbl(virtVal.doubleValue()));
      }
    }
    res.append("\n");
    return res.toString();
  }

  public int totalSenses() {
    int res = 0;
    Enumeration wEnum = words.elements();
    while (wEnum.hasMoreElements()) {
      res += ((Word)wEnum.nextElement()).totalSenses();
    }
    return res;
  }
}
