package NTLSystem.verblearn;

import java.io.*;
import java.util.*;
import NTLSystem.support.*;

/**
 * Representation of a word.  Mainly consists of a set of senses.
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: Word.java,v 1.2 1998/07/31 21:11:32 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Sense
 * @version   $Revision: 1.2 $
 **/
public class Word
    implements Serializable
{

    /**
     * Convenience value: divisor for converting log to ln.
     **/
    public static final double LN_TO_LOG = Math.log(10);

    /**
     * The name of this Word.
     **/
    private String name;

    /**
     * The collection of Senses which make up this Word.
     * Vector<Sense>.
     **/
    private Vector senses;
    
    /**
     * The candidate merges, given a new example.  Orderd by
     * similarity. 
     **/
    private PriorityQueue candidateMerges;  // ordered by similarity

    /** Number of examples incorporated since last merge. **/
    private int backlog; 

    /**
     * A Vector of Fstructs seen over all Senses of this Word.
     * Used in computing likelihoods.
     **/
    private Vector examples;

    /**
     * Used to give unique names to senses.
     **/
    private int nextSenseNumber;

    /**
     * The Lexicon which contains this Word.
     **/
    private Lexicon lex;

    /**
     * A no-argument constructor for serialization.
     **/
    public Word()
    { }

    /**
     * Constructor for a Word.
     *
     * @param     newName The 'name' of this Word.  Should be the same
     *            as the string rep of the label.
     * @param     l The Lexicon which contains this Word.
     **/
    public Word(String newName, Lexicon l)
    {
	name = w;
	senses = new Vector();
	candidateMerges = new PriorityQueue();
	examples = new Vector();
	nextSenseNumber = 0;
	lex = l;
    }

    /**
     * Get the name of this Word.
     *
     * @return    The name of this Word.
     **/
    public String getName() {return name;}
    
    /**
     * Set the name of this Word.
     *
     * @param     newName The new name for this Word.
     **/
    public void setName(String newName) {this.name = newName;}

    /**
     * Returns the number of examples which this Word covers.
     **/
    public double numExamples() { return examples.size(); }
    
    /**
     * Returns the Lexicon which contains this Word, or null if this
     * Word is not contained in a Lexicon.
     *
     * @return    The Lexicon which contains this Word.
     **/
    public Lexicon getContainer() { return lex; }

    /**
     * Sets the containing Lexicon for this Word.
     *
     * @param     lex The Lexicon which contains this Word.
     **/
    public void setContainer(Lexicon lex) { container = lex; }
	
    /**
     * Incorporate a new training instance, by adding a new sense for
     * it and possibly performing some merging (returns true iff so).
     * 
     * @param     f The Fstruct describing the action to incorporate.
     **/
    public boolean incorporate(Fstruct f) {
	examples.addElement(f);
	VerbLearn.trainLog().append("Creating new sense " + s);
	// FIXME:
	Sense s = new Sense(name + nextSenseNumber, f, slot.virtuals);
	nextSenseNumber++;
	backlog++;
	senses.addElement(s);
	addNewCandidateMerges(s);
	// batchSize ?= 'k' in "Modeling Embodied Lexical Development"?
	if (backlog >= VerbLearn.context().params().batchSize()) {
	    synch();
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Merges and sets the backlog to 0.
     *
     * @see       mergeUntilLocalMax()
     **/
    public void synch() {
	mergeUntilLocalMax();
	backlog = 0;
    }

    /**
     * Find best merge candidates and perform the merge, repeating
     * until no satisfactory merge candidates are found or the
     * posterior drops. 
     **/
    private void mergeUntilLocalMax() {

	// Determine the posterior for the current Word model:

	double logPrior = logPrior();
	double logLike = logLikelihoodOfDataset();
	double oldLogPost = logPrior + logLike;
	double newLogPost = 0;  // this initialization is only to satisfy compiler
	double rate;
	Sense newSense = null;
	if (VerbLearn.context().params().testRecognition()) {
	    rate  = recognitionRate();
	    String rateStr;
	    if (rate >= 0) {
		rateStr = String.valueOf((int)(rate * 100)) + "%";
	    }
	    else {
		rateStr = "N/A";
	    }
	    VerbLearn.trainLog().append("Starting merging for " + id + 
					" (log prior=" + Fmt.dbl(logPrior) + 
					", log likelihood=" + Fmt.dbl(logLike) +
					", log posterior=" + Fmt.dbl(oldLogPost) + 
					", recog rate=" + rateStr +
					"):\n");
	}
	else {
	    VerbLearn.trainLog().append("Starting merging for " + id + 
					" (log prior=" + Fmt.dbl(logPrior) + 
					", log likelihood=" + Fmt.dbl(logLike) +
					", log posterior=" + Fmt.dbl(oldLogPost) + 
					"):\n");
	}

	// Repeatedly find and commit merges until no acceptable merge can
	// be found:

	while (true) {

	    CandidateMerge cand;
	    boolean commitMerge = false;
	    int failedTries = 0;
	    Vector putBackList = new Vector();  // Vector<CandidateMerge>

	    // Try out merges until there are no remaining "plausible"
	    // (adequate similarity) merges on the queue, or MaxTries
	    // tentative merges have failed to raise the posterior:

	    while (true) {

		// Pick the highest-similarity candidate merge if such exists
		// and see if it's plausible enough:

		cand = (CandidateMerge)candidateMerges.remove();
		if (cand == null ||
		    cand.similarity < VerbLearn.context().params().minMerge()) { 
		    if (senses.size() == 1) {
			VerbLearn.trainLog().append("Stopped merging because only " +
						    "one sense remains.\n");
		    }
		    else {
			VerbLearn.trainLog().append("Stopped merging because no (more) " +
						    "plausible merges were found.\n");
		    }
		    break;
		}

		// Try out the candidate merge:

		newSense = merge(cand);
		logPrior = logPrior();
		logLike = logLikelihoodOfDataset();
		newLogPost = logPrior + logLike;
		if (VerbLearn.context().params().testRecognition()) {
		    rate = recognitionRate();
		    String rateStr;
		    if (rate >= 0) {
			rateStr = String.valueOf((int)(rate * 100)) + "%";
		    } else {
			rateStr = "N/A";
		    }
		    VerbLearn.trainLog().append("New " + id + 
						" log prior=" + Fmt.dbl(logPrior) + 
						", log likelihood=" + Fmt.dbl(logLike) +
						", log posterior=" + Fmt.dbl(newLogPost)+
						", recog rate=" + rateStr +
						".\n");
		} else {
		    VerbLearn.trainLog().append("New " + id + 
						" log prior=" + Fmt.dbl(logPrior) + 
						", log likelihood=" + Fmt.dbl(logLike) +
						", log posterior=" + Fmt.dbl(newLogPost)+
						".\n");
		}

		// If an improvement, indicate the merge should be committed,
		// otherwise undo it and see if we should try again:

		if (newLogPost >= oldLogPost) {
		    commitMerge = true;
		    break;
		}
		else {
		    unmerge(cand, newSense);
		    putBackList.addElement(cand);
		    failedTries++;
		    if (failedTries == VerbLearn.context().params().maxTries()) {
			VerbLearn.trainLog()
			    .append("Stopped merging because the " + 
				    (VerbLearn.context().params().maxTries() > 1 ? 
				     (String.valueOf(VerbLearn.context().params()
						     .maxTries()) + " most plausible merges ") : 
				     "most plausible merge ") +
				    "reduced the posterior.\n");
			break;
		    }
		}
	    }

	    // Restore bum merges since they might (?) work out in the future:
	    for (int i = 0; i < putBackList.size(); i++) {
		CandidateMerge cm = (CandidateMerge)putBackList.elementAt(i);
		candidateMerges.insert(cm, cm.similarity);
	    }

	    // Commit the Word model to the most-recent merge if indicated,
	    // else terminate the outer merging loop:

	    if (commitMerge) {
		purgeCandidateMerges(cand);
		addNewCandidateMerges(newSense);
		oldLogPost = newLogPost;
		if (failedTries > 0) {
		    VerbLearn.trainLog().append("*** AN EXAMPLE OF KEEPING A " +
						"NON-TOP-RANKED MERGE ***");
		}
	    } else {
		break;
	    }
	}
    }

    /**
     * Returns new sense to allow caller to undo merge with
     * unmerge(). 
     *
     * @param     cand The CandidateMerge to perform.
     * @return    The new Sense resulting from the merge operation.
     **/
    private Sense merge(CandidateMerge cand) {
	senses.removeElement(cand.s1);
	senses.removeElement(cand.s2);
	// Note: uses the current virtual samples for the new sense:
	Sense newSense = new Sense(cand.s1,
				   cand.s2,
				   name + nextSenseNumber,
				   this);
	nextSenseNumber++;
	senses.addElement(newSense);
	VerbLearn.trainLog().append("Merging " + cand.s1.id() + " and " + 
				    cand.s2.id() + " (similarity " + 
				    Fmt.dbl(cand.similarity) + ") to form " +
				    newSense);
	return newSense;
    }
    
    /**
     * Rolls back a merge operation.  Intended for use when a merge
     * lowers the posterior of the model. 
     *
     * @param     cand The CandidateMerge to undo.
     * @param     newSense The new Sense which should be removed, as
     *            it's creation lowered the posterior.
     **/
    private void unmerge(CandidateMerge cand, Sense newSense) {
	senses.removeElement(newSense);
	senses.addElement(cand.s1);
	senses.addElement(cand.s2);
	VerbLearn.trainLog().append("Posterior decreased; removing " +
				    newSense.id() + " and restoring " + 
				    cand.s1.id() + " and " + cand.s2.id() + ".\n");
    }

    /**
     * Adds to candidateMerges all the new potential merges involving
     * a new Sense.
     *
     * @param     newSense The new Sense to add potential merges of.
     **/
    private void addNewCandidateMerges(Sense newSense) {
	Enumeration sEnum = senses.elements();
	while (sEnum.hasMoreElements()) {
	    Sense s = (Sense)sEnum.nextElement();
	    if (s != newSense) {
		CandidateMerge cand = new CandidateMerge(newSense, s, 
							 newSense.similarityTo(s));
		candidateMerges.insert(cand, cand.similarity);
	    }
	} // end while over enumeration of Senses
    }

    /**
     * Removes from candidateMerges all potential merges involving
     * either Sense contained in 'merged'.
     *
     * @param     merged The merged Senses.
     **/
    private void purgeCandidateMerges(CandidateMerge merged) {
	// THIS IS REALLY GROSS AND SHOULD BE REPLACED BY SOMETHING EFFICIENT:
	PriorityQueue pq = new PriorityQueue();
	while (!candidateMerges.isEmpty()) {
	    CandidateMerge cand = (CandidateMerge)candidateMerges.remove();
	    if (cand.s1 != merged.s1 && cand.s1 != merged.s2 &&
		cand.s2 != merged.s1 && cand.s2 != merged.s2) {
		pq.insert(cand, cand.similarity);
	    }
	}
	candidateMerges = pq;
    }

    /**
     * Prior on current word model.
     * <P>
     * Currently we use an exponentially decreasing function of
     * the number of word senses.
     * <P>
     * PERHAPS USE ANDREAS' IDEA TO INCREASE PRIOR WEIGHT DURING
     * TRAINING TO AVOID EARLY OVER-MERGING?
     * <P>
     * Return the base-10 log of this prior.
     *
     * @return    The prior of the Word.
     **/
    private double logPrior() {
	// FIXME:
	// double w = getLexicon().getScenarioSet().getParams().
	//            getParam( Param.modelPriorWeight );
	double w = VerbLearn.context().params().modelPriorWeight();
	double normalization = Math.exp(w) * (1 - Math.exp(-w));
	double prior = normalization * Math.exp(-w * senses.size());
	return Math.log(prior) / LN_TO_LOG;
    }

    /**
     * Likelihood of the set of examples given the word model.
     * Uses log base 10.
     *
     * @return    The log [base 10] of the likelihood of the examples
     *            covered by this Word.
     **/
    private double logLikelihoodOfDataset() {
	double res = 0;
	Enumeration exEnum = examples.elements();
	while (exEnum.hasMoreElements()) {
	    res = res + Math.log(likelihood((Fstruct)exEnum.nextElement()))
		/ LN_TO_LOG;
	}
	return res;
    }

    /**
     * Sums the likelihoods across all senses, weighted by frequency
     * of each sense.
     * <P>
     * TBD: SHOULD FREQUENCY BE USED ONLY IF USEWORDFREQUENCIES IS
     * ENABLED? 
     * 
     * @param     f The Fstruct to compute likelihood against.
     * @return    The likelihood across all Senses, weighted by Sense
     *            frequency. 
     **/
    private double likelihood(Fstruct f) {
	double res = 0;
	Enumeration sEnum = senses.elements();
	while (sEnum.hasMoreElements()) {
	    Sense s = (Sense)sEnum.nextElement();
	    //     |------------ weight -----------|-> Sense frequency
	    res += (s.numExamples() / numExamples()) * s.likelihood(f);
	}
	return res;
    }

    /**
     * Test recognition results ONLY for this word, and report the
     * fraction which are correctly labelled.
     * If there are NO examples to test, a NEGATIVE NUMBER is
     * returned. 
     *
     * @return    The rate of recognition, or a negative number if
     *            there are no examples to recognize against.
     **/
    public double recognitionRate()
    {
	// THIS CODE IS ADAPTED FROM VERBLEARN.RECOGNIZE()
	int numCorrect = 0;
	int numTested = 0;
	Enumeration scEnum = VerbLearn.context().dataset().recognizeEnum();
	while (scEnum.hasMoreElements()) {
	    Scenario sc =
		VerbLearn.context().scenarios().get( (String)scEnum.nextElement());
	    // ONLY TEST EXAMPLES OF THIS CURRENT WORD
	    //
	    // FIXME_jsled:
	    // if (sc.isLabeled() &&
	    //     sc.label().equals( this.getLabel() ) {
	    if (sc.isLabelled() &&
		sc.label().getWord(slot.id()).equals(id)) {
		String actual = VerbLearn.context().lexicon().label(sc.finalLink())
		    .toString();
		numTested++;
		if (actual.equals(sc.label().toString())) {
		    numCorrect++;
		}
	    }
	}
	if (numTested > 0) {
	    return (double)numCorrect / (double)numTested;
	} else {
	    return -1;
	}
    }

  // Return the closest matching Sense to 'f'.  The 'posterior' field
  // includes a prior reflecting the relative frequency of the sense
  // relative to the 'divisor' parameter, which should be the Slot's
  // total number of examples (for labelling) or the Word's total
  // number of examples (for obeying).  Also note: the label field of
  // the return value is unused.
  // THIS ROUTINE AND ITS RETURN TYPE ARE GETTING TOO OVERLOADED.
  public CandidateSense bestSenseFor(Fstruct f, double divisor) {
    CandidateSense leading = new CandidateSense();
    leading.posterior = -1;
    Enumeration sEnum = senses.elements();
    while (sEnum.hasMoreElements()) {
      Sense s = (Sense)sEnum.nextElement();
      double posterior;
      if (VerbLearn.context().params().useWordFrequencies()) {
	posterior = (s.numEx() / divisor) * s.likelihood(f);
      }
      else {
	posterior = s.likelihood(f);
      }
      if (posterior > leading.posterior) {
	leading.sense = s;
	leading.posterior = posterior;
      }
    }
    return leading;
  }

  // Return posteriors of all word senses as a Vector<CandidateSense>.
  // 'divisor' should be the Slot's total number of examples.
  public Vector allSensePosteriors(Fstruct f, double divisor) {
    Vector res = new Vector();  // Vector<CandidateSense>
    Enumeration sEnum = senses.elements();
    while (sEnum.hasMoreElements()) {
      Sense s = (Sense)sEnum.nextElement();
      double pr;
      if (VerbLearn.context().params().useWordFrequencies()) {
	pr = (s.numEx() / divisor);
      }
      else {
	pr = 1;
      }
      double li = s.likelihood(f);
      res.addElement(new CandidateSense(null, s, pr, li, pr * li));
    }
    return res;
  }

  public Enumeration getsenses() { return senses.elements(); }

  // NC note: this is function to obviate by using a browser
  public String toString() {
    StringBuffer res = new StringBuffer();
    res.append("Model for ").append(id).append(" (").append(senses.size())
      .append(" sense");
    if (senses.size() > 1) res.append("s");
    res.append("):\n");
    Enumeration sEnum = senses.elements();
    while (sEnum.hasMoreElements()) {
      res.append("\n").append(sEnum.nextElement());
    }
    return res.toString();
  }

  public int totalSenses() { return senses.size(); }
}
