package NTLSystem.verblearn;

import java.io.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * A Sense of a Word is a collection of Features which describe that
 * Sense.  Therefore, it is an Fstruct, but has some extended methods
 * for handling VerbLearn-specific things [likelihood, similarityTo,
 * etc].  As well, a Sense is usually created by taking the union of
 * the Fstructs which describe a Scenario, and modifying the names of
 * the Features to prevent name clashes.  For instance, if a Scenario
 * contains N Fstructs, each with M Features, the Sense will contain
 * the set Features <TT>{ Fstruct1_Name.Feature1_Name,
 * Fstruct1_Name.Feature2_Name, ... , Fstruct1_Name.FeatureM_Name,
 * Fstruct2_Name.Feature1_Name, Fstruct2_Name.Feature2_Name, ... ,
 * Fstruct2_Name.FeatureM_Name, ... , FstructN_Name.FeatureM_Name
 * }</TT>, and will have cardinality NM.  Of course, not all Fstructs
 * in a Scenario are required to have the same number of Features...
 * <P>
 * From "Modeling Embodied Lexical Development":
 * <BLOCKQUOTE>
 * "Each Sense of a Verb is represented in the model by an Fstruct
 * whose values for each Feature are probability distributions.
 * <BR>[...]<BR>
 * With this scheme it is necessary to allow multiple senses for each
 * verb [Word], since capturing the meaning in a single Sense would in
 * some cases require overly broad distrubtions which obscure details
 * needed to carry out the corresponding action."
 * </BLOCKQUOTE>
 * <P>
 * In VerbLearn, all Fstructs are probabilistic, and contain
 * DiscreteDists.  This may change in the future, but for now that's
 * the way it is.  These assumptions are reflected in the code
 * [casting for DiscreteDist, assuming ProbDist values on Features,
 * etc] 
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: Sense.java,v 1.7 1998/10/16 01:59:42 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @author    Josh Sled [jsled@icsi.berkeley.edu]
 * @see       Word
 * @see       Lexicon
 * @version   $Revision: 1.7 $
 **/
public class Sense
    extends Fstruct
    implements Serializable
{

    // {{{ ATTRIBUTES

    /** Serialization stream unique identifier.  Recompute [with
	serialver] when this class changes. **/
    static final long serialVersionUID = -7462934535655350856L;

    /** The Word which contains this Sense. **/
    private Word container;

    /** Number of examples covered by this Sense. **/
    private int numEx;

    // }}} end attributes

    // {{{ CONSTRUCTORS
    /**
     * Turn an instance into an initial Sense.  Features are copied
     * and converted into VirtualSampleDiscreteDists from the
     * DiscreteDists they are assumed to be.  Properties are
     * <STRONG>not</STRONG> copied.
     *
     * @param     name Name of the Sense.
     * @param     f Fstruct on which Sense is based. (Based on a Scenario)
     * @param     w The Word which contains this Sense.
     **/
    public Sense(String name, Fstruct f, Word w) {
	// Set the name, copy the Features, and leave the Properties
	// alone. 
	super( name, VerbLearn.getFeatureDict() );
	container = w;
	numEx = 1;
	try {
	    Enumeration fNames = f.getAllFeatureNames().elements();
	    Vector features = f.getFeatures( fNames );
	    // to hold the features after we've cloned them.  Then, we
	    // can add them as a whole with putFeatures.
	    Vector toAdd = new Vector( features.size() );
	    for (int i=0 ; i<features.size(); i++) {
		Feature feat =
		    (Feature)((Feature)features.elementAt(i)).clone();
		double vSamps =
		    getWord().getLexicon().
		    getVirtualSamples( feat.getName() );
		
		VirtualSampleDiscreteDist vsdd = null;
		if (feat.getValue() != null) {
		    vsdd =
			new VirtualSampleDiscreteDist((DiscreteDist)feat.getValue(),
						      vSamps );
		} else {
		    // If it doesn't have a value, make an empty distribution
		    // of the correct size.
		    DiscreteFeatureDescriptor dfd =
			(DiscreteFeatureDescriptor)feat.getDescriptor();
		    vsdd =
			new VirtualSampleDiscreteDist( dfd.getValueCount().intValue(),
						       vSamps );
		}
		feat.setValue( vsdd );
		toAdd.addElement( feat  );
	    } // end for over i
	    putFeatures( toAdd.elements() );
	} catch (NoSuchFeatureException nsfe) {
	    // Shouldn't happen, as we just got the names from 'f'.
	    throw new InternalError( nsfe.getMessage() );
	} catch (NTLException ntle) {
	    // Shouldn't happen, as we're giving it Strings like it
	    // expects. 
	    throw new InternalError( ntle.getMessage() );
	} // end try-catch
    }

    /**
     * Merge two Senses together into a single Sense, using the given
     * virtuals.
     *
     * @param     s1 The first Sense to merge.
     * @param     s2 The second Sense to merge.
     * @param     newSenseName The name of the merged Sense
     * @param     w The Word which contains this new Sense.
     **/
    public Sense(Sense s1,
		 Sense s2,
		 String newSenseName,
		 Word w)
    {
	super( newSenseName, s1.getDictionary() );
	container = w;
	numEx = s1.numExamples() + s2.numExamples();
	// Go through and merge the Features.  We assume that the
	// Senses have the same Features, because they should.  This
	// is the Lexicon, after all...
	Vector fNames = s1.getAllFeatureNames();
	Vector features1 = null;
	Vector features2 = null;
	try {
	    features1 = s1.getFeatures( fNames.elements() );
	    features2 = s2.getFeatures( fNames.elements() );
	    if (features1.size() != features2.size()) {
		// FIXME: error
		System.out.println("features1.size=" + features1.size() + 
				   " != features2.size=" + features2.size() );
	    }
	    Vector toAdd = new Vector( features1.size() );
	    Feature f, f1, f2;
	    VirtualSampleDiscreteDist vsdd1, vsdd2;
	    for (int i=0; i<features1.size(); i++) {
		f1 = (Feature)features1.elementAt( i );
		f2 = (Feature)features2.elementAt( i );
		f = (Feature)f1.clone();
		vsdd1 = (VirtualSampleDiscreteDist)f1.getValue();
		vsdd2 = (VirtualSampleDiscreteDist)f2.getValue();
		double vSamps =
		    getWord().getLexicon().
		    getVirtualSamples( f1.getName() );
		if ( (vsdd1 != null) && (vsdd2 == null) ) {
		    vsdd1 = (VirtualSampleDiscreteDist)vsdd1.clone();
		    vsdd1.setVirtualSamples( vSamps );
		    f.setValue( vsdd1 );
		} else if ( (vsdd1 == null) && (vsdd2 != null) ) {
		    vsdd2 = (VirtualSampleDiscreteDist)vsdd2.clone();
		    vsdd2.setVirtualSamples( vSamps );
		    f.setValue( vsdd2 );
		} else if ( (vsdd1 == null) && (vsdd2 == null) ) {
		    f.setValue( (VirtualSampleDiscreteDist)null );
		} else {
		    f.setValue(new VirtualSampleDiscreteDist(vsdd1,
							     vsdd2,
							     vSamps));
		}
		toAdd.addElement( f );
	    } // end for over features [i]
	    putFeatures( toAdd.elements() );
	} catch ( Exception e ) {
	    // We _really_ don't expect any exceptions here, since
	    // we're adding Features.
	    System.out.println( e.getMessage() );
	    throw new InternalError( e.getMessage() );
	} // end try-catch
    }

    // }}} Constructors

    // {{{ Methods

    /**
     * Access method for number of examples this Sense covers.
     *
     * @return    The number of examples this Sense covers.
     **/
    public int numExamples() { return numEx; }

    /**
     * The the Word which contains this Sense.
     *
     * @return    The Word which contains this Sense, or null if this
     *            Sense is not contained in a Word.
     **/
    public Word getWord() { return container; }

    /**
     * Computes the likelihood of this Sense given another Sense.
     * <P>
     * This is used by model-merging in NTLSystem.verblearn.Lexicon.
     * Therefore, there are the follwing restrictions:
     * <UL>
     * <LI>The Features in both Fstructs [Sense] must be Probabilistic
     * and Discrete, and therefore contain DiscreteDists.
     * <LI>Each Feature in 'f' must have a maximum-likelihood value
     * for the selected value [GUI sets these to 1.0 for the selected
     * value, 0.0 for the others].
     * </UL>
     * <P>
     * <STRONG>NOTE:</STRONG><BR>
     * If you intend to use this method, look it over and make sure
     * that it will work with whatever you're trying to do.  The code
     * here is fairly VerbLearn-specific... the likelihood of two
     * random Fstructs is undefined [or else this method would be in
     * NTLSystem.Fstruct]. 
     *
     * @param     f The Sense to use to calculate the likelihood. 
     * @return    The likelihood of this Sense given the Feature
     *            values of the other Sense.
     **/
    public double likelihood(Fstruct f) {
	double res = 1;
	double li;
	Vector fNames = f.getAllFeatureNames();
	Feature thisF, fF;

	for (int i=0; i<fNames.size(); i++) {
	    String fname = (String)fNames.elementAt( i );
	    try {
		thisF = getFeature( fname );
		fF = f.getFeature( fname );
		// Make sure that the FeatureDescriptors are the same,
		// the Feature[s] is [are] Probabilistic, and
		// the ProbDist types are DiscreteDist.
		//
		// If not, silently ignore the Feature.
		/*		if ( thisF.getDescriptor().
		     equals( fF.getDescriptor() ) ) {		*/

		    // f.getFeature( fname ).getValue().getMLValue()
		    // is equivalent to the 'f.get(fname)' method from
		    // the old verblearn: In that, only the selected
		    // feature value was stored.  In the new setup,
		    // the selected feature value has probability of
		    // 1.0 [the rest are 0.0], and 'mode()' will
		    // return it.
		    if ( (thisF.getValue() != null) &&
			 (fF.getValue() != null)       ) {
			li = ((VirtualSampleDiscreteDist)thisF.
			     getValue()).getProb(
			       ((DiscreteDist)fF.getValue()).
			       mode().intValue() ); 
			res *= li;
			    
		    }
		    /*} */
	    } catch (NoSuchFeatureException nsfe) {
		// This isn't a good thing... in VerbLearn, it
		// shouldn't happen, since the Fstruct given should
		// have identical Features to the Senses in the
		// Lexicon... 
		//
		// FIXME: How does this affect the likelihood?
				
		System.out.println( "Ignorning missing Feature: " +
				    nsfe.getMessage() );
	    } catch (NTLException ntle) {
		// This shouldn't happen, since we know we're giving
		// it Strings.
		throw new InternalError( ntle.getMessage() );
	    } // end try-catch
	} // end for over i
	return res;
    }

    /**
     * A heuristic measure of the similarity between two Fstructs with
     * probabilistic Features. Currently we multiply the similarities
     * between the distributions for each feature but this may or may
     * not be the best combination function.
     * <P>
     * This was taken from dbailey's old VerbLearn
     * [java/Lexicon.java::PFstruct], and is more than likely
     * <STRONG>not</STRONG> correct outside that context.  If you
     * intend to use this method, look it over and make sure it does
     * what you want.
     *
     * @param     f The Sense [Fstruct] to compute the similarity to.
     * @return    A double representing the similarity between the
     *            Senses.
     **/
    public double similarityTo(Sense f) {
	double res = 1;
	Vector fNames = f.getAllFeatureNames();
	Feature thisF, fF;

	for (int i=0; i<fNames.size(); i++) {
	    String fname = (String)fNames.elementAt( i );
	    try {
		thisF = getFeature( fname );
		fF = f.getFeature( fname );
		// This makes certain assumptions about the values
		// which the Features contain.  However, in the
		// VerbLearn context, they are justified. 
		if ( (thisF.getValue() != null) &&
		     (fF.getValue() != null)       ) {
		    res *= ((VirtualSampleDiscreteDist)thisF.
			    getValue()).
			similarityTo(((VirtualSampleDiscreteDist)fF.
				      getValue()) );
		}
	    } catch (NoSuchFeatureException nsfe) {
		// This isn't a good thing... in VerbLearn, it
		// shouldn't happen, since the Fstruct given should
		// have identical Features to the Senses in the
		// Lexicon... 
		//
		// FIXME: How does this affect the similarity?
		// Probably negatively....
		// For now, silently ignore missing Features.
		System.out.println( "Ignorning missing Feature: " +
				    nsfe.getMessage() );
	    } catch (NTLException ntle) {
		// This shouldn't happen, since we know we're giving
		// it Strings.
		throw new InternalError( ntle.getMessage() );
	    } // end try-catch
	} // end for over i
	return res;
    }

    /**
     * Deconstructs a Sense [which is the union of Fstructs, with
     * Features differentiated by naming convention] into an array of
     * Fstructs with their original names.  These Fstructs contain
     * clones of the Features in the Sense.  The names of these
     * Features are set to 'null', resulting in the original Feature
     * name to be used.  The Fstructs are returned in the same order
     * as the template Fstructs in the ScenarioSet. 
     * 
     * @param     fd The FeatureDictionary to use for the new
     *            Fstructs.
     * @return    An array of Fstructs, with their corresponding
     *            Features from this Sense.
     **/
    public Fstruct[] splitFstructs(FeatureDictionary fd)
    {
	// This maps Fstruct names to Feature vectors to be used for
	// creating Fstructs to be returned.
	StringSortedHashtable fstructs = new StringSortedHashtable();
	Vector templFstructs =
	    getWord().getLexicon().getScenarioSet().getTemplateFstructs();
	Fstruct toRet[] = new Fstruct[ templFstructs.size() ];
	int fstructCount = 0;
	Enumeration featNames = getAllFeatureNames().elements();
	Vector myFeatures = null;
	try {
	    myFeatures = getFeatures(featNames);
	} catch (NTLException ntle) {
	    // We expect neither NTLException [because we're giving
	    // Strings we just got from ourself] nor
	    // NoSuchFeatureException [because we're asking for
	    // features we just told this method we have].
	    throw new InternalError( ntle.getMessage() );
	}

	// Go through our Features and break them into groups based on
	// the first part of their name [before the dot ... the
	// Fstruct name].
	Feature f;
	String fName; // feature name
	String fsName, // fstruct name [substring of fName]
	    featName; // feature name [substring of fName]
	int dotIdx; // index of the '.' between the two above parts.
	Vector featsToAdd; // reference to the Vector [stored in
	                   // fstructs] of Features we're splitting
                           // up. 
	for (int i=0; i<myFeatures.size(); i++) {
	    //   Get the Feature name
	    f = (Feature)myFeatures.elementAt(i);
	    fName = (String)f.getName();
	    //   Cut it at the first '.'
	    dotIdx = fName.indexOf( '.' );
	    fsName = fName.substring( 0, dotIdx );
	    featName = fName.substring( dotIdx+1 );
	    featsToAdd = (Vector)fstructs.get( fsName );
	    if (featsToAdd == null) {
		featsToAdd = new Vector();
		fstructCount++;
		fstructs.put( fsName, featsToAdd );
	    }
	    f = (Feature)f.clone();
	    // Set the feature name to null to use the original
	    // FeatureDescriptor name.
	    f.setName( null );
	    featsToAdd.addElement( f );
	} // end for over Features

	// Sanity check/assertion
	if (fstructCount != toRet.length) {
	    throw new InternalError("Mismatch between Fstructs in Sense " +
				    fstructCount +
				    " and template Fstruct length " +
				    toRet.length);
	}

	// Now that the Features are split into groups, create
	// Fstructs for original Fstruct [in the same order as the
	// template Fstructs] with their Features in it.
	for (int j=0; j<templFstructs.size(); j++) {
	    fsName = ((Fstruct)templFstructs.elementAt(j)).getName();
	    toRet[j] = new Fstruct( fsName, fd );
	    // Get the correct features from their group we created
	    // above. 
	    try {
		toRet[j].putFeatures( ((Vector)fstructs.get( fsName )).
				      elements() );
	    } catch (NTLException ntle) {
		// We don't expect either NTLException [because we're
		// indeed giving it Features], nor
		// NoSuchFeatureException [because we're giving it the
		// correct FeatureDictionary].
		throw new InternalError( ntle.getMessage() );
	    } // end try-catch
	} // end for over j [template Fstructs]

	return toRet;
    }

    public String toString()
    {
	StringBuffer toRet = new StringBuffer( 500 );
	toRet.append("Sense {\n");
	toRet.append("  numExamples: ").append( numEx );
	String superRet = super.toString();
	superRet = superRet.substring( superRet.indexOf("{")+1 );
	toRet.append( superRet );
	return toRet.toString();
    }

    // }}} Methods
}
